from __future__ import annotations

import posixpath
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    TypeVar,
    overload,
)

from typing_extensions import Self

if TYPE_CHECKING:
    import requests

    from .context import Context
    from .urls import Url


@dataclass(frozen=True)
class ResourceParameters:
    """Shared parameter object for resources.

    Attributes
    ----------
    session: requests.Session
        A `requests.Session` object. Provides cookie persistence, connection-pooling, and
        configuration.
    url: str
        The Connect API base URL (e.g., https://connect.example.com/__api__)
    """

    session: requests.Session
    url: Url


class Resource(dict):
    def __init__(self, /, params: ResourceParameters, **kwargs):
        self.params = params
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            warnings.warn(
                f"Accessing the field '{name}' via attribute is deprecated and will be removed in v1.0.0. "
                f"Please use __getitem__ (e.g., {self.__class__.__name__.lower()}['{name}']) for field access instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return self[name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)


class Resources:
    def __init__(self, params: ResourceParameters) -> None:
        self.params = params


class Active(ABC, Resource):
    def __init__(self, ctx: Context, path: str, /, **attributes):
        """A dict abstraction for any HTTP endpoint that returns a singular resource.

        Extends the `Resource` class and provides additional functionality for via the session context and an optional parent resource.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        path : str
            The HTTP path component for the resource endpoint
        **attributes : dict
            Resource attributes passed
        """
        params = ResourceParameters(ctx.session, ctx.url)
        super().__init__(params, **attributes)
        self._ctx = ctx
        self._path = path


T = TypeVar("T", bound="Active")
"""A type variable that is bound to the `Active` class"""


class ActiveSequence(ABC, Generic[T], Sequence[T]):
    """A sequence for any HTTP GET endpoint that returns a collection."""

    _cache: Optional[List[T]]

    def __init__(self, ctx: Context, path: str, uid: str = "guid"):
        """A sequence abstraction for any HTTP GET endpoint that returns a collection.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        path : str
            The HTTP path component for the collection endpoint
        uid : str, optional
            The field name of that uniquely identifiers an instance of T, by default "guid"
        """
        super().__init__()
        self._ctx = ctx
        self._path = path
        self._uid = uid
        self._cache: Optional[List[T]] = None

    @abstractmethod
    def _create_instance(self, path: str, /, **kwargs: Any) -> T:
        """Create an instance of 'T'."""
        raise NotImplementedError()

    def fetch(self, **conditions: Any) -> Iterable[T]:
        """Fetch the collection.

        Fetches the collection directly from Connect. This operation does not effect the cache state.

        Returns
        -------
        List[T]
        """
        endpoint = self._ctx.url + self._path
        response = self._ctx.session.get(endpoint, params=conditions)
        results = response.json()
        return [self._to_instance(result) for result in results]

    def reload(self) -> Self:
        """Reloads the collection from Connect.

        Returns
        -------
        Self
        """
        self._cache = None
        return self

    def _to_instance(self, result: dict) -> T:
        """Converts a result into an instance of T."""
        uid = result[self._uid]
        path = posixpath.join(self._path, uid)
        return self._create_instance(path, **result)

    @property
    def _data(self) -> List[T]:
        """Get the collection.

        Fetches the collection from Connect and caches the result. Subsequent invocations return the cached results unless the cache is explicitly reset.

        Returns
        -------
        List[T]

        See Also
        --------
        cached
        reload
        """
        if self._cache is None:
            self._cache = list(self.fetch())
        return self._cache

    @overload
    def __getitem__(self, index: int) -> T: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[T]: ...

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


class ActiveFinderMethods(ActiveSequence[T]):
    """Finder methods.

    Provides various finder methods for locating records in any endpoint supporting HTTP GET requests.
    """

    def find(self, uid) -> T:
        """
        Find a record by its unique identifier.

        Fetches the record from Connect by it's identifier.

        Parameters
        ----------
        uid : Any
            The unique identifier of the record.

        Returns
        -------
        T
        """
        endpoint = self._ctx.url + self._path + uid
        response = self._ctx.session.get(endpoint)
        result = response.json()
        return self._to_instance(result)

    def find_by(self, **conditions: Any) -> T | None:
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        **conditions : Any

        Returns
        -------
        Optional[T]
            The first record matching the conditions, or `None` if no match is found.
        """
        collection = self.fetch(**conditions)
        return next((v for v in collection if v.items() >= conditions.items()), None)



import pytest
import requests
import responses
from responses import matchers

from posit.connect.client import Client
from posit.connect.content import ContentItem, ContentItemRepository
from posit.connect.context import Context
from posit.connect.resources import ResourceParameters
from posit.connect.urls import Url

from .api import load_mock, load_mock_dict


class TestContentItemGetContentOwner:
    @responses.activate
    def test_owner(self):
        mock_content = load_mock_dict("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json")
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=mock_content,
        )

        mock_user_get = responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        c = Client("https://connect.example", "12345")
        item = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        owner = item.owner
        assert owner["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"

        # load a second time, assert tha owner is loaded from cached result
        owner = item.owner
        assert owner["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"
        assert mock_user_get.call_count == 1


class TestContentItemDelete:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_delete = responses.delete(f"https://connect.example/__api__/v1/content/{guid}")

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(guid)

        # invoke
        content.delete()

        # assert
        assert mock_delete.call_count == 1


class TestContentDeploy:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        task_id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_content_deploy = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/deploy",
            match=[matchers.json_params_matcher({"bundle_id": None})],
            json={"task_id": task_id},
        )

        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{task_id}",
            json=load_mock(f"v1/tasks/{task_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(content_guid)

        # invoke
        task = content.deploy()

        # assert
        assert task["id"] == task_id
        assert mock_content_get.call_count == 1
        assert mock_content_deploy.call_count == 1
        assert mock_tasks_get.call_count == 1


class TestContentUpdate:
    @responses.activate
    def test_update(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )
        con = Client("https://connect.example", "12345")
        content = con.content.get(guid)
        assert content["guid"] == guid

        new_name = "New Name"
        fake_content = load_mock_dict(f"v1/content/{guid}.json")
        fake_content.update(name=new_name)
        responses.patch(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=fake_content,
        )

        content.update(name=new_name)
        assert content["name"] == new_name


class TestContentCreate:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_content_item = load_mock_dict(f"v1/content/{guid}.json")

        # behavior
        responses.post(
            "https://connect.example/__api__/v1/content",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[matchers.json_params_matcher({"name": fake_content_item["name"]})],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example/")

        # invoke
        fake_name = fake_content_item["name"]
        assert isinstance(fake_name, str)
        content_item = client.content.create(name=fake_name)

        # assert
        assert content_item["name"] == fake_content_item["name"]


class TestContentsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find()

        # assert
        assert mock_get.call_count == 1
        assert len(content) == 3
        assert content[0]["name"] == "team-admin-dashboard"
        assert content[1]["name"] == "Performance-Data-1671216053560"
        assert content[2]["name"] == "My-Streamlit-app"

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_list(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags,owner"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include=["tags", "owner"])

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsFindBy:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find_by(name="team-admin-dashboard")

        # assert
        assert mock_get.call_count == 1
        assert content
        assert content["name"] == "team-admin-dashboard"

    @responses.activate
    def test_miss(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find_by(name="does-not-exist")

        # assert
        assert mock_get.call_count == 1
        assert content is None


class TestContentsFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one()

        #  assert
        assert mock_get.call_count == 1
        assert content_item

    @responses.activate
    def test_owner_guid(self):
        # data
        owner_guid = "a01792e3-2e67-402e-99af-be04a48da074"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one(owner_guid=owner_guid)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item["owner_guid"] == owner_guid

    @responses.activate
    def test_name(self):
        # data
        name = "team-admin-dashboard"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one(name=name)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item["name"] == name

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find_one(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find_one(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsGet:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )
        con = Client("https://connect.example", "12345")
        get_one = con.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        assert get_one["name"] == "Performance-Data-1671216053560"


class TestContentsCount:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.content.count()
        assert count == 3


class TestRender:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        get_content = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        get_variants = responses.get(
            f"https://connect.example.com/__api__/applications/{guid}/variants",
            json=load_mock(f"applications/{guid}/variants.json"),
        )

        post_render = responses.post(
            "https://connect.example.com/__api__/variants/6627/render",
            json=load_mock("variants/6627/render.json"),
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        task = content.render()

        # assert
        assert task is not None
        assert get_content.call_count == 1
        assert patch_content.call_count == 1
        assert get_variants.call_count == 1
        assert post_render.call_count == 1

    @responses.activate
    def test_app_mode_is_other(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="other")

        # behavior
        responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(ValueError):
            content.render()

    @responses.activate
    def test_missing_default(self):
        responses.get(
            "https://connect.example.com/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.patch(
            "https://connect.example.com/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example.com/__api__/applications/f2f37341-e21d-3d80-c698-a935ad614066/variants",
            json=[],
        )

        c = Client("https://connect.example.com", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        with pytest.raises(RuntimeError):
            content.render()


class TestRestart:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="api")

        # behavior
        mock_get_content = mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_environment = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        )

        mock_get_content_page = responses.get(
            f"https://connect.example.com/content/{guid}",
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        task = content.restart()

        # assert
        assert task is None
        assert mock_get_content.call_count == 1
        assert mock_patch_content.call_count == 1
        assert mock_patch_environment.call_count == 2
        assert mock_get_content_page.call_count == 1

    @responses.activate
    def test_app_mode_is_other(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="other")

        # behavior
        mock_get_content = mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(ValueError):
            content.restart()

        # assert
        assert mock_get_content.call_count == 1
        assert mock_patch_content.call_count == 1


class TestContentRepository:
    @property
    def base_url(self):
        return "http://connect.example"

    @property
    def content_guid(self):
        return "f2f37341-e21d-3d80-c698-a935ad614066"

    @property
    def content_item(self):
        return ContentItem(self.params, guid=self.content_guid)

    @property
    def endpoint(self):
        return f"{self.base_url}/__api__/v1/content/{self.content_guid}/repository"

    @property
    def ctx(self):
        return Context(requests.Session(), Url(self.base_url))

    @property
    def params(self):
        return ResourceParameters(self.ctx.session, self.ctx.url)

    def mock_repository_info(self):
        content_item = self.content_item

        mock_get = responses.get(
            self.endpoint,
            json=load_mock_dict(f"v1/content/{self.content_guid}/repository.json"),
        )
        repository_info = content_item.repository

        assert isinstance(repository_info, ContentItemRepository)
        assert mock_get.call_count == 1

        return repository_info

    @responses.activate
    def test_repository_getter_returns_repository(self):
        # Performs assertions in helper property method
        self.mock_repository_info()

    @responses.activate
    def test_repository_update(self):
        repository_info = self.mock_repository_info()

        mock_patch = responses.patch(
            self.endpoint,
            json=load_mock_dict(f"v1/content/{self.content_guid}/repository_patch.json"),
        )
        new_repository_info = repository_info.update(branch="testing-main")
        assert mock_patch.call_count == 1

        for key, value in repository_info.items():
            if key == "branch":
                assert new_repository_info[key] == "testing-main"
            else:
                assert new_repository_info[key] == value

    @responses.activate
    def test_repository_delete(self):
        repository_info = self.mock_repository_info()

        mock_delete = responses.delete(self.endpoint)
        repository_info.destroy()

        assert mock_delete.call_count == 1



from pathlib import Path

import pytest
from packaging import version

from posit import connect

from . import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION < version.parse("2024.10.0-dev"),
    reason="Packages API unavailable",
)
class TestPackages:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(name=cls.__name__)
        path = Path("../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = cls.content.bundles.create(str(path))
        task = bundle.deploy()
        task.wait_for()

    @classmethod
    def teardown_class(cls):
        cls.content.delete()

    def test(self):
        assert self.client.packages
        assert self.content.packages

    def test_find_by(self):
        package = self.client.packages.find_by(name="flask")
        assert package
        assert package["name"] == "flask"

        package = self.content.packages.find_by(name="flask")
        assert package
        assert package["name"] == "flask"



import pytest

from posit.connect.errors import ClientError


def test():
    error_code = 0
    error_message = "error"
    http_status = 404
    http_message = "Error"
    with pytest.raises(
        ClientError,
        match='{"error_code": 0, "error_message": "error", "http_status": 404, "http_message": "Error", "payload": null}',
    ):
        raise ClientError(
            error_code=error_code,
            error_message=error_message,
            http_status=http_status,
            http_message=http_message,
        )


def test_payload_is_str():
    error_code = 0
    error_message = "error"
    http_status = 404
    http_message = "Error"
    payload = "This is an error payload"
    with pytest.raises(
        ClientError,
        match='{"error_code": 0, "error_message": "error", "http_status": 404, "http_message": "Error", "payload": "This is an error payload"}',
    ):
        raise ClientError(
            error_code=error_code,
            error_message=error_message,
            http_status=http_status,
            http_message=http_message,
            payload=payload,
        )


def test_payload_is_dict():
    error_code = 0
    error_message = "error"
    http_status = 404
    http_message = "Error"
    payload = {"message": "This is an error payload"}
    with pytest.raises(
        ClientError,
        match='{"error_code": 0, "error_message": "error", "http_status": 404, "http_message": "Error", "payload": {"message": "This is an error payload"}}',
    ):
        raise ClientError(
            error_code=error_code,
            error_message=error_message,
            http_status=http_status,
            http_message=http_message,
            payload=payload,
        )



"""OAuth session resources."""

from typing import List, Optional, overload

from ..resources import Resource, Resources


class Session(Resource):
    """OAuth session resource."""

    def delete(self) -> None:
        path = f"v1/oauth/sessions/{self['guid']}"
        url = self.params.url + path
        self.params.session.delete(url)


class Sessions(Resources):
    @overload
    def find(
        self,
        *,
        all: Optional[bool] = ...,
    ) -> List[Session]: ...

    @overload
    def find(self, **kwargs) -> List[Session]: ...

    def find(self, **kwargs) -> List[Session]:
        url = self.params.url + "v1/oauth/sessions"
        response = self.params.session.get(url, params=kwargs)
        results = response.json()
        return [Session(self.params, **result) for result in results]

    def get(self, guid: str) -> Session:
        """Get an OAuth session.

        Parameters
        ----------
        guid: str

        Returns
        -------
        Session
        """
        path = f"v1/oauth/sessions/{guid}"
        url = self.params.url + path
        response = self.params.session.get(url)
        return Session(self.params, **response.json())



# TODO-barret-future; Piecemeal migrate everything to leverage `ApiDictEndpoint` and `ApiListEndpoint` classes.
# TODO-barret-future; Merge any trailing behavior of `Active` or `ActiveList` into the new classes.

from __future__ import annotations

import itertools
import posixpath
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Generator, Generic, Optional, TypeVar, cast, overload

from ._api_call import ApiCallMixin, get_api
from ._json import Jsonifiable, JsonifiableDict, ResponseAttrs

if TYPE_CHECKING:
    from .context import Context


# Design Notes:
# * Perform API calls on property retrieval. e.g. `my_content.repository`
# * Dictionary endpoints: Retrieve all attributes during init unless provided
# * List endpoints: Do not retrieve until `.fetch()` is called directly. Avoids cache invalidation issues.
#   * While slower, all ApiListEndpoint helper methods should `.fetch()` on demand.
# * Only expose methods needed for `ReadOnlyDict`.
#   * Ex: When inheriting from `dict`, we'd need to shut down `update`, `pop`, etc.
# * Use `ApiContextProtocol` to ensure that the class has the necessary attributes for API calls.
#    * Inherit from `ApiCallMixin` to add all helper methods for API calls.
# * Classes should write the `path` only once within its init method.
#    * Through regular interactions, the path should only be written once.

# When making a new class,
# * Use a class to define the parameters and their types
#    * If attaching the type info class to the parent class, start with `_`. E.g.: `ContentItemRepository._Attrs`
# * Document all attributes like normal
#    * When the time comes that there are multiple attribute types, we can use overloads with full documentation and unpacking of type info class for each overload method.
# * Inherit from `ApiDictEndpoint` or `ApiListEndpoint` as needed
#    * Init signature should be `def __init__(self, ctx: Context, path: str, /, **attrs: Jsonifiable) -> None:`


# This class should not have typing about the class keys as that would fix the class's typing. If
# for some reason, we _know_ the keys are fixed (as we've moved on to a higher version), we can add
# `Generic[AttrsT]` to the class.
class ReadOnlyDict(Mapping):
    _attrs: ResponseAttrs
    """Resource attributes passed."""

    def __init__(self, attrs: ResponseAttrs) -> None:
        """
        A read-only dict abstraction for any HTTP endpoint that returns a singular resource.

        Parameters
        ----------
        attrs : dict
            Resource attributes passed
        """
        super().__init__()
        self._attrs = attrs

    def get(self, key: str, default: Any = None) -> Any:
        return self._attrs.get(key, default)

    def __getitem__(self, key: str) -> Any:
        return self._attrs[key]

    def __setitem__(self, key: str, value: Any) -> None:
        raise NotImplementedError(
            "Resource attributes are locked. "
            "To retrieve updated values, please retrieve the parent object again."
        )

    def __len__(self) -> int:
        return self._attrs.__len__()

    def __iter__(self):
        return self._attrs.__iter__()

    def __contains__(self, key: object) -> bool:
        return self._attrs.__contains__(key)

    def __repr__(self) -> str:
        return repr(self._attrs)

    def __str__(self) -> str:
        return str(self._attrs)

    def keys(self):
        return self._attrs.keys()

    def values(self):
        return self._attrs.values()

    def items(self):
        return self._attrs.items()


class ApiDictEndpoint(ApiCallMixin, ReadOnlyDict):
    _ctx: Context
    """The context object containing the session and URL for API interactions."""
    _path: str
    """The HTTP path component for the resource endpoint."""

    def _get_api(self, *path) -> JsonifiableDict | None:
        super()._get_api(*path)

    def __init__(
        self,
        ctx: Context,
        path: str,
        get_data: Optional[bool] = None,
        /,
        **attrs: Jsonifiable,
    ) -> None:
        """
        A dict abstraction for any HTTP endpoint that returns a singular resource.

        Adds helper methods to interact with the API with reduced boilerplate.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        path : str
            The HTTP path component for the resource endpoint
        get_data : Optional[bool]
            If `True`, fetch the API and set the attributes from the response. If `False`, only set
            the provided attributes. If `None` [default], fetch the API if no attributes are
            provided.
        attrs : dict
            Resource attributes passed
        """
        # If no attributes are provided, fetch the API and set the attributes from the response
        if get_data is None:
            get_data = len(attrs) == 0

        # If we should get data, fetch the API and set the attributes from the response
        if get_data:
            init_attrs: Jsonifiable = get_api(ctx, path)
            init_attrs_dict = cast(ResponseAttrs, init_attrs)
            # Overwrite the initial attributes with `attrs`: e.g. {'key': value} | {'content_guid': '123'}
            init_attrs_dict.update(attrs)
            attrs = init_attrs_dict

        super().__init__(attrs)
        self._ctx = ctx
        self._path = path


T = TypeVar("T", bound="ReadOnlyDict")
"""A type variable that is bound to the `Active` class"""


class ApiListEndpoint(ApiCallMixin, Generic[T], ABC, object):
    """A HTTP GET endpoint that can fetch a collection."""

    def __init__(self, *, ctx: Context, path: str, uid_key: str = "guid") -> None:
        """A sequence abstraction for any HTTP GET endpoint that returns a collection.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        path : str
            The HTTP path component for the collection endpoint
        uid_key : str, optional
            The field name of that uniquely identifiers an instance of T, by default "guid"
        """
        super().__init__()
        self._ctx = ctx
        self._path = path
        self._uid_key = uid_key

    @abstractmethod
    def _create_instance(self, path: str, /, **kwargs: Any) -> T:
        """Create an instance of 'T'."""
        raise NotImplementedError()

    def fetch(self) -> Generator[T, None, None]:
        """Fetch the collection.

        Fetches the collection directly from Connect. This operation does not effect the cache state.

        Returns
        -------
        List[T]
        """
        results: Jsonifiable = self._get_api()
        results_list = cast(list[JsonifiableDict], results)
        for result in results_list:
            yield self._to_instance(result)

    def __iter__(self) -> Generator[T, None, None]:
        return self.fetch()

    def _to_instance(self, result: dict) -> T:
        """Converts a result into an instance of T."""
        uid = result[self._uid_key]
        path = posixpath.join(self._path, uid)
        return self._create_instance(path, **result)

    @overload
    def __getitem__(self, index: int) -> T: ...

    @overload
    def __getitem__(self, index: slice) -> Generator[T, None, None]: ...

    def __getitem__(self, index: int | slice) -> T | Generator[T, None, None]:
        if isinstance(index, slice):
            results = itertools.islice(self.fetch(), index.start, index.stop, index.step)
            for result in results:
                yield result
        else:
            return list(itertools.islice(self.fetch(), index, index + 1))[0]

    # def __len__(self) -> int:
    #     return len(self.fetch())

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # Jobs - 123 items
        return repr(
            f"{self.__class__.__name__} - { len(list(self.fetch())) } items - {self._path}"
        )

    def find(self, uid: str) -> T | None:
        """
        Find a record by its unique identifier.

        Fetches the record from Connect by it's identifier.

        Parameters
        ----------
        uid : str
            The unique identifier of the record.

        Returns
        -------
        :
            Single instance of T if found, else None
        """
        result: Jsonifiable = self._get_api(uid)
        result_obj = cast(JsonifiableDict, result)

        return self._to_instance(result_obj)

    def find_by(self, **conditions: Any) -> T | None:
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        **conditions : Any

        Returns
        -------
        T
            The first record matching the conditions, or `None` if no match is found.
        """
        results = self.fetch()

        conditions_items = conditions.items()

        # Get the first item of the generator that matches the conditions
        # If no item is found, return None
        return next(
            (
                # Return result
                result
                # Iterate through `results` generator
                for result in results
                # If all `conditions`'s key/values are found in `result`'s key/values...
                if result.items() >= conditions_items
            ),
            None,
        )



import pytest
import responses

from posit.connect.client import Client

from .api import load_mock  # type: ignore


class TestPackagesMixin:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/packages",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066/packages.json"),
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")

        content._ctx.version = None
        assert len(content.packages) == 1


class TestPackagesFind:
    @responses.activate
    def test(self):
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        with pytest.raises(NotImplementedError):
            c.packages.find("posit")


class TestPackagesFindBy:
    @responses.activate
    def test(self):
        mock_get = responses.get(
            "https://connect.example/__api__/v1/packages",
            json=load_mock("v1/packages.json"),
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        package = c.packages.find_by(name="posit")
        assert package
        assert package["name"] == "posit"
        assert mock_get.call_count == 1



import pytest

from posit.connect._api import ReadOnlyDict


class TestApiEndpoint:
    def test_read_only(self):
        obj = ReadOnlyDict({})

        assert len(obj) == 0

        assert obj.get("foo", "bar") == "bar"

        with pytest.raises(NotImplementedError):
            obj["foo"] = "baz"

        eq_obj = ReadOnlyDict({"foo": "bar", "a": 1})
        assert eq_obj == {"foo": "bar", "a": 1}



"""Client connection for Posit Connect."""

from __future__ import annotations

from typing import overload

from requests import Response, Session

from . import hooks, me
from .auth import Auth
from .config import Config
from .content import Content
from .context import Context, ContextManager, requires
from .groups import Groups
from .metrics import Metrics
from .oauth import OAuth
from .packages import Packages
from .resources import ResourceParameters
from .tasks import Tasks
from .users import User, Users
from .vanities import Vanities


class Client(ContextManager):
    """
    Client connection for Posit Connect.

    This class provides an interface to interact with the Posit Connect API,
    allowing for authentication, resource management, and data retrieval.

    Parameters
    ----------
    api_key : str, optional
        API key for authentication
    url : str, optional
        Sever API URL

    Attributes
    ----------
    content: Content
        Content resource.
    me: User
        Connect user resource.
    metrics: Metrics
        Metrics resource.
    oauth: OAuth
        OAuth resource.
    tasks: Tasks
        Tasks resource.
    users: Users
        Users resource.
    version: str
        Server version.
    """

    @overload
    def __init__(self) -> None:
        """Initialize a Client instance.

        Creates a client instance using credentials read from the environment.

        Environment Variables
        ---------------------
        CONNECT_SERVER - The Connect server URL.
        CONNECT_API_KEY - The API key credential for client authentication.

        Examples
        --------
        Client()
        """

    @overload
    def __init__(self, url: str) -> None:
        """Initialize a Client instance.

        Creates a client instance using a provided URL and API key credential read from the environment.

        Environment Variables
        ---------------------
        CONNECT_API_KEY - The API key credential for client authentication.

        Parameters
        ----------
        url : str
            The Connect server URL.

        Examples
        --------
        Client("https://connect.example.com)
        """

    @overload
    def __init__(self, url: str, api_key: str) -> None:
        """Initialize a Client instance.

        Parameters
        ----------
        url : str
            The Connect server URL.
        api_key : str
            The API key credential for client authentication.

        Examples
        --------
        >>> Client("https://connect.example.com", abcdefghijklmnopqrstuvwxyz012345")
        """

    def __init__(self, *args, **kwargs) -> None:
        """Initialize a Client instance.

        Environment Variables
        ---------------------
        CONNECT_SERVER - The Connect server URL.
        CONNECT_API_KEY - The API key credential for client authentication.

        Parameters
        ----------
        *args
            Variable length argument list. Can accept:
            - (url: str)
                url: str
                    The Connect server URL.
            - (url: str, api_key: str)
                url: str
                    The Connect server URL.
                api_key: str
                    The API key credential for client authentication.

        **kwargs
            Keyword arguments. Can include 'url' and 'api_key'.

        Examples
        --------
        >>> Client()
        >>> Client("https://connect.example.com")
        >>> Client("https://connect.example.com", abcdefghijklmnopqrstuvwxyz012345")
        >>> Client(api_key=""abcdefghijklmnopqrstuvwxyz012345", url="https://connect.example.com")
        """
        api_key = None
        url = None
        if len(args) == 1 and isinstance(args[0], str):
            url = args[0]
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            url = args[0]
            api_key = args[1]
        else:
            if "api_key" in kwargs and isinstance(kwargs["api_key"], str):
                api_key = kwargs["api_key"]
            if "url" in kwargs and isinstance(kwargs["url"], str):
                url = kwargs["url"]

        self.cfg = Config(api_key=api_key, url=url)
        session = Session()
        session.auth = Auth(config=self.cfg)
        session.hooks["response"].append(hooks.check_for_deprecation_header)
        session.hooks["response"].append(hooks.handle_errors)
        self.session = session
        self.resource_params = ResourceParameters(session, self.cfg.url)
        self._ctx = Context(self.session, self.cfg.url)

    @property
    def version(self) -> str | None:
        """
        The server version.

        Returns
        -------
        str
            The version of the Posit Connect server.
        """
        return self._ctx.version

    @property
    def me(self) -> User:
        """
        The connected user.

        Returns
        -------
        User
            The currently authenticated user.
        """
        return me.get(self.resource_params)

    @property
    def groups(self) -> Groups:
        """The groups resource interface.

        Returns
        -------
        Groups
            The groups resource interface.
        """
        return Groups(self.resource_params)

    @property
    def tasks(self) -> Tasks:
        """
        The tasks resource interface.

        Returns
        -------
        tasks.Tasks
            The tasks resource instance.
        """
        return Tasks(self.resource_params)

    @property
    def users(self) -> Users:
        """
        The users resource interface.

        Returns
        -------
        Users
            The users resource instance.
        """
        return Users(self.resource_params)

    @property
    def content(self) -> Content:
        """
        The content resource interface.

        Returns
        -------
        Content
            The content resource instance.
        """
        return Content(self.resource_params)

    @property
    def metrics(self) -> Metrics:
        """
        The Metrics API interface.

        The Metrics API is designed for capturing, retrieving, and managing
        quantitative measurements of Connect interactions. It is commonly used
        for monitoring and analyzing system performance, user behavior, and
        business processes. This API facilitates real-time data collection and
        accessibility, enabling organizations to make informed decisions based
        on key performance indicators (KPIs).

        Returns
        -------
        Metrics
            The metrics API instance.

        Examples
        --------
        >>> from posit import connect
        >>> client = connect.Client()
        >>> content_guid = "2243770d-ace0-4782-87f9-fe2aeca14fc8"
        >>> events = client.metrics.usage.find(content_guid=content_guid)
        >>> len(events)
        24
        """
        return Metrics(self.resource_params)

    @property
    @requires(version="2024.08.0")
    def oauth(self) -> OAuth:
        """
        The OAuth API interface.

        Returns
        -------
        OAuth
            The oauth API instance.
        """
        return OAuth(self.resource_params, self.cfg.api_key)

    @property
    @requires(version="2024.10.0-dev")
    def packages(self) -> Packages:
        return Packages(self._ctx, "v1/packages")

    @property
    def vanities(self) -> Vanities:
        return Vanities(self.resource_params)

    def __del__(self):
        """Close the session when the Client instance is deleted."""
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def __enter__(self):
        """Enter method for using the client as a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Close the session if it exists.

        Parameters
        ----------
        exc_type : type
            The type of the exception raised (if any).
        exc_value : Exception
            The exception instance raised (if any).
        exc_tb : traceback
            The traceback for the exception raised (if any).
        """
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def request(self, method: str, path: str, **kwargs) -> Response:
        """
        Send an HTTP request.

        A facade for [](`requests.request`) configured for the target server.

        Parameters
        ----------
        method : str
            The HTTP method to use for the request.
        path : str
            Appended to the url object attribute.
        **kwargs
            Additional keyword arguments passed to [](`requests.request`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.request(method, url, **kwargs)

    def get(self, path: str, **kwargs) -> Response:
        """
        Send a GET request.

        A facade for [](`requests.get`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.get`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.get(url, **kwargs)

    def post(self, path: str, **kwargs) -> Response:
        """
        Send a POST request.

        A facade for [](`requests.post`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.post`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.post(url, **kwargs)

    def put(self, path: str, **kwargs) -> Response:
        """
        Send a PUT request.

        A facade for [](`requests.put`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.put`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.put(url, **kwargs)

    def patch(self, path: str, **kwargs) -> Response:
        """
        Send a PATCH request.

        A facade for [](`requests.patch`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.patch`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.patch(url, **kwargs)

    def delete(self, path: str, **kwargs) -> Response:
        """
        Send a DELETE request.

        A facade for [](`requests.delete`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.delete`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.delete(url, **kwargs)



import random
import uuid

import requests
import responses
from responses import matchers

from posit.connect.permissions import Permission, Permissions
from posit.connect.resources import ResourceParameters
from posit.connect.urls import Url

from .api import load_mock, load_mock_dict, load_mock_list


class TestPermissionDelete:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        fake_permission = load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json")
        permission = Permission(params, **fake_permission)

        # invoke
        permission.delete()

        # assert
        assert mock_delete.call_count == 1


class TestPermissionUpdate:
    @responses.activate
    def test_request_shape(self):
        # test data
        uid = random.randint(0, 100)
        content_guid = str(uuid.uuid4())
        principal_guid = str(uuid.uuid4())
        principal_type = "principal_type"
        role = "role"
        extraneous = "extraneous"

        # define api behavior
        responses.put(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json={
                # doesn't matter for this test
            },
            match=[
                # assertion
                matchers.json_params_matcher(
                    {
                        # validate that initial permission fields are set
                        "principal_guid": principal_guid,
                        "principal_type": principal_type,
                        "role": role,
                        # validate that arguments passed to update are set
                        "extraneous": extraneous,
                    },
                ),
            ],
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permission = Permission(
            params,
            id=uid,
            content_guid=content_guid,
            principal_guid=principal_guid,
            principal_type=principal_type,
            role=role,
        )

        # invoke
        # assertion occurs in match above
        permission.update(extraneous=extraneous)

    @responses.activate
    def test_role_update(self):
        # test data
        old_role = "owner"
        new_role = "viewer"

        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permission = load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json")
        fake_permission.update(role=new_role)

        # define api behavior
        uid = random.randint(0, 100)
        content_guid = str(uuid.uuid4())
        responses.put(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json=fake_permission,
            match=[
                matchers.json_params_matcher(
                    {
                        "principal_guid": None,
                        "principal_type": None,
                        "role": new_role,
                    },
                ),
            ],
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permission = Permission(params, id=uid, content_guid=content_guid, role=old_role)

        # assert role change with respect to api response
        assert permission["role"] == old_role
        permission.update(role=new_role)
        assert permission["role"] == new_role


class TestPermissionsCount:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock_list(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permissions = Permissions(params, content_guid=content_guid)

        # invoke
        count = permissions.count()

        # assert response
        assert count == len(fake_permissions)


class TestPermissionsCreate:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        principal_guid = str(uuid.uuid4())
        principal_type = "user"
        role = "owner"
        fake_permission = {
            **load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json"),
            "principal_guid": principal_guid,
            "principal_type": principal_type,
            "role": role,
        }

        # behavior
        responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permission,
            match=[
                matchers.json_params_matcher(
                    {
                        "principal_guid": principal_guid,
                        "principal_type": principal_type,
                        "role": role,
                    },
                ),
            ],
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permissions = Permissions(params, content_guid=content_guid)

        # invoke
        permission = permissions.create(
            principal_guid=principal_guid,
            principal_type=principal_type,
            role=role,
        )

        # assert
        assert permission == fake_permission


class TestPermissionsFind:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permissions = Permissions(params, content_guid=content_guid)

        # invoke
        permissions = permissions.find()

        # assert response
        assert permissions == fake_permissions


class TestPermissionsFindOne:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock_list(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permissions = Permissions(params, content_guid=content_guid)

        # invoke
        permission = permissions.find_one()

        # assert response
        assert permission == fake_permissions[0]


class TestPermissionsGet:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permission = load_mock(f"v1/content/{content_guid}/permissions/{uid}.json")

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json=fake_permission,
        )

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))
        permissions = Permissions(params, content_guid=content_guid)

        # invoke
        permission = permissions.get(uid)

        # assert
        assert permission == fake_permission



"""Bundle resources."""

from __future__ import annotations

import io
from typing import List

from . import resources, tasks


class BundleMetadata(resources.Resource):
    pass


class Bundle(resources.Resource):
    @property
    def metadata(self) -> BundleMetadata:
        return BundleMetadata(self.params, **self.get("metadata", {}))

    def delete(self) -> None:
        """Delete the bundle."""
        path = f"v1/content/{self['content_guid']}/bundles/{self['id']}"
        url = self.params.url + path
        self.params.session.delete(url)

    def deploy(self) -> tasks.Task:
        """Deploy the bundle.

        Spawns an asynchronous task, which activates the bundle.

        Returns
        -------
        tasks.Task
            The task for the deployment.

        Examples
        --------
        >>> task = bundle.deploy()
        >>> task.wait_for()
        None
        """
        path = f"v1/content/{self['content_guid']}/deploy"
        url = self.params.url + path
        response = self.params.session.post(url, json={"bundle_id": self["id"]})
        result = response.json()
        ts = tasks.Tasks(self.params)
        return ts.get(result["task_id"])

    def download(self, output: io.BufferedWriter | str) -> None:
        """Download a bundle.

        Download a bundle to a file or memory.

        Parameters
        ----------
        output : io.BufferedWriter or str
            An io.BufferedWriter instance or a str representing a relative or absolute path.

        Raises
        ------
        TypeError
            If the output is not of type `io.BufferedWriter` or `str`.

        Examples
        --------
        Write to a file.
        >>> bundle.download("bundle.tar.gz")
        None

        Write to an io.BufferedWriter.
        >>> with open('bundle.tar.gz', 'wb') as file:
        >>>     bundle.download(file)
        None
        """
        if not isinstance(output, (io.BufferedWriter, str)):
            raise TypeError(
                f"download() expected argument type 'io.BufferedWriter` or 'str', but got '{type(output).__name__}'",
            )

        path = f"v1/content/{self['content_guid']}/bundles/{self['id']}/download"
        url = self.params.url + path
        response = self.params.session.get(url, stream=True)
        if isinstance(output, io.BufferedWriter):
            for chunk in response.iter_content():
                output.write(chunk)
        elif isinstance(output, str):
            with open(output, "wb") as file:
                for chunk in response.iter_content():
                    file.write(chunk)


class Bundles(resources.Resources):
    """Bundles resource.

    Parameters
    ----------
    config : config.Config
        Configuration object.
    session : requests.Session
        HTTP session object.
    content_guid : str
        Content GUID associated with the bundles.

    Attributes
    ----------
    content_guid: str
        Content GUID associated with the bundles.
    """

    def __init__(
        self,
        params: resources.ResourceParameters,
        content_guid: str,
    ) -> None:
        super().__init__(params)
        self.content_guid = content_guid

    def create(self, archive: io.BufferedReader | bytes | str) -> Bundle:
        """
        Create a bundle.

        Create a bundle from a file or memory.

        Parameters
        ----------
        archive : io.BufferedReader, bytes, or str
            Archive for bundle creation. A 'str' type assumes a relative or absolute filepath.

        Returns
        -------
        Bundle
            The created bundle.

        Raises
        ------
        TypeError
            If the input is not of type `io.BufferedReader`, `bytes`, or `str`.

        Examples
        --------
        Create a bundle from io.BufferedReader
        >>> with open('bundle.tar.gz', 'rb') as file:
        >>>     bundle.create(file)
        None

        Create a bundle from bytes.
        >>> with open('bundle.tar.gz', 'rb') as file:
        >>>     data: bytes = file.read()
        >>>     bundle.create(data)
        None

        Create a bundle from pathname.
        >>> bundle.create("bundle.tar.gz")
        None
        """
        if isinstance(archive, (io.BufferedReader, bytes)):
            data = archive
        elif isinstance(archive, str):
            with open(archive, "rb") as file:
                data = file.read()
        else:
            raise TypeError(
                f"create() expected argument type 'io.BufferedReader', 'bytes', or 'str', but got '{type(archive).__name__}'",
            )

        path = f"v1/content/{self.content_guid}/bundles"
        url = self.params.url + path
        response = self.params.session.post(url, data=data)
        result = response.json()
        return Bundle(self.params, **result)

    def find(self) -> List[Bundle]:
        """Find all bundles.

        Returns
        -------
        list of Bundle
            List of all found bundles.
        """
        path = f"v1/content/{self.content_guid}/bundles"
        url = self.params.url + path
        response = self.params.session.get(url)
        results = response.json()
        return [Bundle(self.params, **result) for result in results]

    def find_one(self) -> Bundle | None:
        """Find a bundle.

        Returns
        -------
        Bundle | None
            The first found bundle | None if no bundles are found.
        """
        bundles = self.find()
        return next(iter(bundles), None)

    def get(self, uid: str) -> Bundle:
        """Get a bundle.

        Parameters
        ----------
        uid : str
            Identifier of the bundle to retrieve.

        Returns
        -------
        Bundle
            The bundle with the specified ID.
        """
        path = f"v1/content/{self.content_guid}/bundles/{uid}"
        url = self.params.url + path
        response = self.params.session.get(url)
        result = response.json()
        return Bundle(self.params, **result)



"""Content resources."""

from __future__ import annotations

import posixpath
import time
from posixpath import dirname
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Literal,
    Optional,
    cast,
    overload,
)

from typing_extensions import NotRequired, Required, TypedDict, Unpack

from . import tasks
from ._api import ApiDictEndpoint, JsonifiableDict
from .bundles import Bundles
from .context import Context
from .env import EnvVars
from .errors import ClientError
from .jobs import JobsMixin
from .oauth.associations import ContentItemAssociations
from .packages import ContentPackagesMixin as PackagesMixin
from .permissions import Permissions
from .resources import Resource, ResourceParameters, Resources
from .vanities import VanityMixin
from .variants import Variants

if TYPE_CHECKING:
    from .tasks import Task


def _assert_guid(guid: str):
    assert isinstance(guid, str), "Expected 'guid' to be a string"
    assert len(guid) > 0, "Expected 'guid' to be non-empty"


def _assert_content_guid(content_guid: str):
    assert isinstance(content_guid, str), "Expected 'content_guid' to be a string"
    assert len(content_guid) > 0, "Expected 'content_guid' to be non-empty"


class ContentItemRepository(ApiDictEndpoint):
    """
    Content items GitHub repository information.

    See Also
    --------
    * Get info: https://docs.posit.co/connect/api/#get-/v1/content/-guid-/repository
    * Delete info: https://docs.posit.co/connect/api/#delete-/v1/content/-guid-/repository
    * Update info: https://docs.posit.co/connect/api/#patch-/v1/content/-guid-/repository
    """

    class _Attrs(TypedDict, total=False):
        repository: str
        """URL for the repository."""
        branch: NotRequired[str]
        """The tracked Git branch."""
        directory: NotRequired[str]
        """Directory containing the content."""
        polling: NotRequired[bool]
        """Indicates that the Git repository is regularly polled."""

    def __init__(
        self,
        ctx: Context,
        /,
        *,
        content_guid: str,
        # By default, the `attrs` will be retrieved from the API if no `attrs` are supplied.
        **attrs: Unpack[ContentItemRepository._Attrs],
    ) -> None:
        """Content items GitHub repository information.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        content_guid : str
            The unique identifier of the content item.
        **attrs : ContentItemRepository._Attrs
            Attributes for the content item repository. If not supplied, the attributes will be
            retrieved from the API upon initialization
        """
        _assert_content_guid(content_guid)

        path = self._api_path(content_guid)
        # Only fetch data if `attrs` are not supplied
        get_data = len(attrs) == 0
        super().__init__(ctx, path, get_data, **{"content_guid": content_guid, **attrs})

    @classmethod
    def _api_path(cls, content_guid: str) -> str:
        return f"v1/content/{content_guid}/repository"

    @classmethod
    def _create(
        cls,
        ctx: Context,
        content_guid: str,
        **attrs: Unpack[ContentItemRepository._Attrs],
    ) -> ContentItemRepository:
        from ._api_call import put_api

        result = put_api(ctx, cls._api_path(content_guid), json=cast(JsonifiableDict, attrs))

        return ContentItemRepository(
            ctx,
            content_guid=content_guid,
            **result,  # pyright: ignore[reportCallIssue]
        )

    def destroy(self) -> None:
        """
        Delete the content's git repository location.

        See Also
        --------
        * https://docs.posit.co/connect/api/#delete-/v1/content/-guid-/repository
        """
        self._delete_api()

    def update(
        self,
        # *,
        **attrs: Unpack[ContentItemRepository._Attrs],
    ) -> ContentItemRepository:
        """Update the content's repository.

        Parameters
        ----------
        repository: str, optional
            URL for the repository. Default is None.
        branch: str, optional
            The tracked Git branch. Default is 'main'.
        directory: str, optional
            Directory containing the content. Default is '.'
        polling: bool, optional
            Indicates that the Git repository is regularly polled. Default is False.

        Returns
        -------
        None

        See Also
        --------
        * https://docs.posit.co/connect/api/#patch-/v1/content/-guid-/repository
        """
        result = self._patch_api(json=cast(JsonifiableDict, dict(attrs)))
        return ContentItemRepository(
            self._ctx,
            content_guid=self["content_guid"],
            **result,  # pyright: ignore[reportCallIssue]
        )


class ContentItemOAuth(Resource):
    def __init__(self, params: ResourceParameters, content_guid: str) -> None:
        super().__init__(params)
        self["content_guid"] = content_guid

    @property
    def associations(self) -> ContentItemAssociations:
        return ContentItemAssociations(self.params, content_guid=self["content_guid"])


class ContentItemOwner(Resource):
    pass


class ContentItem(JobsMixin, PackagesMixin, VanityMixin, Resource):
    class _AttrsBase(TypedDict, total=False):
        # # `name` will be set by other _Attrs classes
        # name: str

        # Content Metadata
        title: NotRequired[str]
        description: NotRequired[str]
        access_type: NotRequired[Literal["all", "acl", "logged_in"]]
        # Timeout Settings
        connection_timeout: NotRequired[int]
        read_timeout: NotRequired[int]
        init_timeout: NotRequired[int]
        idle_timeout: NotRequired[int]
        # Process and Resource Limits
        max_processes: NotRequired[int]
        min_processes: NotRequired[int]
        max_conns_per_process: NotRequired[int]
        load_factor: NotRequired[float]
        cpu_request: NotRequired[float]
        cpu_limit: NotRequired[float]
        memory_request: NotRequired[int]
        memory_limit: NotRequired[int]
        amd_gpu_limit: NotRequired[int]
        nvidia_gpu_limit: NotRequired[int]
        # Execution Settings
        run_as: NotRequired[str]
        run_as_current_user: NotRequired[bool]
        default_image_name: NotRequired[str]
        default_r_environment_management: NotRequired[bool]
        default_py_environment_management: NotRequired[bool]
        service_account_name: NotRequired[str]

    class _AttrsNotRequired(_AttrsBase):
        name: NotRequired[str]
        owner_guid: NotRequired[str]

    class _Attrs(_AttrsBase):
        name: Required[str]
        owner_guid: NotRequired[str]

    class _AttrsCreate(_AttrsBase):
        name: NotRequired[str]
        # owner_guid is not supported

    @overload
    def __init__(
        self,
        /,
        params: ResourceParameters,
        guid: str,
    ) -> None: ...

    @overload
    def __init__(
        self,
        /,
        params: ResourceParameters,
        guid: str,
        **kwargs: Unpack[ContentItem._Attrs],
    ) -> None: ...

    def __init__(
        self,
        /,
        params: ResourceParameters,
        guid: str,
        **kwargs: Unpack[ContentItem._AttrsNotRequired],
    ) -> None:
        _assert_guid(guid)

        ctx = Context(params.session, params.url)
        path = f"v1/content/{guid}"
        super().__init__(ctx, path, guid=guid, **kwargs)

    def __getitem__(self, key: Any) -> Any:
        v = super().__getitem__(key)
        if key == "owner" and isinstance(v, dict):
            return ContentItemOwner(params=self.params, **v)
        return v

    @property
    def oauth(self) -> ContentItemOAuth:
        return ContentItemOAuth(self.params, content_guid=self["guid"])

    @property
    def repository(self) -> ContentItemRepository | None:
        try:
            return ContentItemRepository(self._ctx, content_guid=self["guid"])
        except ClientError:
            return None

    def create_repository(
        self,
        **attrs: Unpack[ContentItemRepository._Attrs],
    ) -> ContentItemRepository:
        """Create repository.

        Parameters
        ----------
        repository : str
            URL for the respository.
        branch : str, optional
            The tracked Git branch. Default is 'main'.
        directory : str, optional
            Directory containing the content. Default is '.'.
        polling : bool, optional
            Indicates that the Git repository is regularly polled. Default is False.

        Returns
        -------
        ContentItemRepository
        """
        return ContentItemRepository._create(self._ctx, self["guid"], **attrs)

    def delete(self) -> None:
        """Delete the content item."""
        path = f"v1/content/{self['guid']}"
        url = self.params.url + path
        self.params.session.delete(url)

    def deploy(self) -> tasks.Task:
        """Deploy the content.

        Spawns an asynchronous task, which activates the latest bundle.

        Returns
        -------
        tasks.Task
            The task for the deployment.

        Examples
        --------
        >>> task = content.deploy()
        >>> task.wait_for()
        None
        """
        path = f"v1/content/{self['guid']}/deploy"
        url = self.params.url + path
        response = self.params.session.post(url, json={"bundle_id": None})
        result = response.json()
        ts = tasks.Tasks(self.params)
        return ts.get(result["task_id"])

    def render(self) -> Task:
        """Render the content.

        Submit a render request to the server for the content. After submission, the server executes an asynchronous process to render the content. This is useful when content is dependent on external information, such as a dataset.

        See Also
        --------
        restart

        Examples
        --------
        >>> render()
        """
        self.update()  # pyright: ignore[reportCallIssue]

        if self.is_rendered:
            variants = self._variants.find()
            variants = [variant for variant in variants if variant["is_default"]]
            if len(variants) != 1:
                raise RuntimeError(
                    f"Found {len(variants)} default variants. Expected 1. Without a single default variant, the content cannot be refreshed. This is indicative of a corrupted state.",
                )
            variant = variants[0]
            return variant.render()
        else:
            raise ValueError(
                f"Render not supported for this application mode: {self['app_mode']}. Did you need to use the 'restart()' method instead? Note that some application modes do not support 'render()' or 'restart()'.",
            )

    def restart(self) -> None:
        """Mark for restart.

        Sends a restart request to the server for the content. Once submitted, the server performs an asynchronous process to restart the content. This is particularly useful when the content relies on external information loaded into application memory, such as datasets. Additionally, restarting can help clear memory leaks or reduce excessive memory usage that might build up over time.

        See Also
        --------
        render

        Examples
        --------
        >>> restart()
        """
        self.update()  # pyright: ignore[reportCallIssue]

        if self.is_interactive:
            unix_epoch_in_seconds = str(int(time.time()))
            key = f"_CONNECT_RESTART_TMP_{unix_epoch_in_seconds}"
            self.environment_variables.create(key, unix_epoch_in_seconds)
            self.environment_variables.delete(key)
            # GET via the base Connect URL to force create a new worker thread.
            url = posixpath.join(dirname(self.params.url), f"content/{self['guid']}")
            self.params.session.get(url)
            return None
        else:
            raise ValueError(
                f"Restart not supported for this application mode: {self['app_mode']}. Did you need to use the 'render()' method instead? Note that some application modes do not support 'render()' or 'restart()'.",
            )

    def update(
        self,
        **attrs: Unpack[ContentItem._Attrs],
    ) -> None:
        """Update the content item.

        Parameters
        ----------
        name : str
            URL-friendly identifier. Allows alphanumeric characters, hyphens ("-"), and underscores ("_").
        title : str, optional
            Content title. Default is None.
        description : str, optional
            Content description. Default is None.
        access_type : Literal['all', 'acl', 'logged_in'], optional
            How content manages viewers. Default is 'acl'. Options: 'all', 'logged_in', 'acl'.
        owner_guid : str, optional
            The unique identifier of the user who owns this content item. Default is None.
        connection_timeout : int, optional
            Max seconds without data exchange. Default is None. Falls back to server setting 'Scheduler.ConnectionTimeout'.
        read_timeout : int, optional
            Max seconds without data received. Default is None. Falls back to server setting 'Scheduler.ReadTimeout'.
        init_timeout : int, optional
            Max startup time for interactive apps. Default is None. Falls back to server setting 'Scheduler.InitTimeout'.
        idle_timeout : int, optional
            Max idle time before process termination. Default is None. Falls back to server setting 'Scheduler.IdleTimeout'.
        max_processes : int, optional
            Max concurrent processes allowed. Default is None. Falls back to server setting 'Scheduler.MaxProcesses'.
        min_processes : int, optional
            Min concurrent processes required. Default is None. Falls back to server setting 'Scheduler.MinProcesses'.
        max_conns_per_process : int, optional
            Max client connections per process. Default is None. Falls back to server setting 'Scheduler.MaxConnsPerProcess'.
        load_factor : float, optional
            Aggressiveness in spawning new processes (0.0 - 1.0). Default is None. Falls back to server setting 'Scheduler.LoadFactor'.
        cpu_request : float, optional
            Min CPU units required (1 unit = 1 core). Default is None. Falls back to server setting 'Scheduler.CPURequest'.
        cpu_limit : float, optional
            Max CPU units allowed. Default is None. Falls back to server setting 'Scheduler.CPULimit'.
        memory_request : int, optional
            Min memory (bytes) required. Default is None. Falls back to server setting 'Scheduler.MemoryRequest'.
        memory_limit : int, optional
            Max memory (bytes) allowed. Default is None. Falls back to server setting 'Scheduler.MemoryLimit'.
        amd_gpu_limit : int, optional
            Number of AMD GPUs allocated. Default is None. Falls back to server setting 'Scheduler.AMDGPULimit'.
        nvidia_gpu_limit : int, optional
            Number of NVIDIA GPUs allocated. Default is None. Falls back to server setting 'Scheduler.NvidiaGPULimit'.
        run_as : str, optional
            UNIX user to execute the content. Default is None. Falls back to server setting 'Applications.RunAs'.
        run_as_current_user : bool, optional
            Run process as the visiting user (for app content). Default is False.
        default_image_name : str, optional
            Default image for execution if not defined in the bundle. Default is None.
        default_r_environment_management : bool, optional
            Manage R environment for the content. Default is None.
        default_py_environment_management : bool, optional
            Manage Python environment for the content. Default is None.
        service_account_name : str, optional
            Kubernetes service account name for running content. Default is None.

        Returns
        -------
        None
        """
        url = self.params.url + f"v1/content/{self['guid']}"
        response = self.params.session.patch(url, json=attrs)
        super().update(**response.json())

    # Relationships

    @property
    def bundles(self) -> Bundles:
        return Bundles(self.params, self["guid"])

    @property
    def environment_variables(self) -> EnvVars:
        return EnvVars(self.params, self["guid"])

    @property
    def permissions(self) -> Permissions:
        return Permissions(self.params, self["guid"])

    @property
    def owner(self) -> dict:
        if "owner" not in self:
            # It is possible to get a content item that does not contain owner.
            # "owner" is an optional additional request param.
            # If it's not included, we can retrieve the information by `owner_guid`
            from .users import Users

            self["owner"] = Users(self.params).get(self["owner_guid"])
        return self["owner"]

    @property
    def _variants(self) -> Variants:
        return Variants(self.params, self["guid"])

    @property
    def is_interactive(self) -> bool:
        return self["app_mode"] in {
            "api",
            "jupyter-voila",
            "python-api",
            "python-bokeh",
            "python-dash",
            "python-fastapi",
            "python-shiny",
            "python-streamlit",
            "quarto-shiny",
            "rmd-shiny",
            "shiny",
            "tensorflow-saved-model",
        }

    @property
    def is_rendered(self) -> bool:
        return self["app_mode"] in {
            "rmd-static",
            "jupyter-static",
            "quarto-static",
        }


class Content(Resources):
    """Content resource.

    Parameters
    ----------
    config : Config
        Configuration object.
    session : Session
        Requests session object.
    owner_guid : str, optional
        Content item owner identifier. Filters results to those owned by a specific user (the default is None, which implies not filtering results on owner identifier).
    """

    def __init__(
        self,
        params: ResourceParameters,
        *,
        owner_guid: str | None = None,
    ) -> None:
        super().__init__(params)
        self.owner_guid = owner_guid

    def count(self) -> int:
        """Count the number of content items.

        Returns
        -------
        int
        """
        return len(self.find())

    def create(
        self,
        **attrs: Unpack[ContentItem._AttrsCreate],
    ) -> ContentItem:
        """Create content.

        Parameters
        ----------
        name : str
            URL-friendly identifier. Allows alphanumeric characters, hyphens ("-"), and underscores ("_").
        title : str, optional
            Content title. Default is None.
        description : str, optional
            Content description. Default is None.
        access_type : Literal['all', 'acl', 'logged_in'], optional
            How content manages viewers. Default is 'acl'. Options: 'all', 'logged_in', 'acl'.
        connection_timeout : int, optional
            Max seconds without data exchange. Default is None. Falls back to server setting 'Scheduler.ConnectionTimeout'.
        read_timeout : int, optional
            Max seconds without data received. Default is None. Falls back to server setting 'Scheduler.ReadTimeout'.
        init_timeout : int, optional
            Max startup time for interactive apps. Default is None. Falls back to server setting 'Scheduler.InitTimeout'.
        idle_timeout : int, optional
            Max idle time before process termination. Default is None. Falls back to server setting 'Scheduler.IdleTimeout'.
        max_processes : int, optional
            Max concurrent processes allowed. Default is None. Falls back to server setting 'Scheduler.MaxProcesses'.
        min_processes : int, optional
            Min concurrent processes required. Default is None. Falls back to server setting 'Scheduler.MinProcesses'.
        max_conns_per_process : int, optional
            Max client connections per process. Default is None. Falls back to server setting 'Scheduler.MaxConnsPerProcess'.
        load_factor : float, optional
            Aggressiveness in spawning new processes (0.0 - 1.0). Default is None. Falls back to server setting 'Scheduler.LoadFactor'.
        cpu_request : float, optional
            Min CPU units required (1 unit = 1 core). Default is None. Falls back to server setting 'Scheduler.CPURequest'.
        cpu_limit : float, optional
            Max CPU units allowed. Default is None. Falls back to server setting 'Scheduler.CPULimit'.
        memory_request : int, optional
            Min memory (bytes) required. Default is None. Falls back to server setting 'Scheduler.MemoryRequest'.
        memory_limit : int, optional
            Max memory (bytes) allowed. Default is None. Falls back to server setting 'Scheduler.MemoryLimit'.
        amd_gpu_limit : int, optional
            Number of AMD GPUs allocated. Default is None. Falls back to server setting 'Scheduler.AMDGPULimit'.
        nvidia_gpu_limit : int, optional
            Number of NVIDIA GPUs allocated. Default is None. Falls back to server setting 'Scheduler.NvidiaGPULimit'.
        run_as : str, optional
            UNIX user to execute the content. Default is None. Falls back to server setting 'Applications.RunAs'.
        run_as_current_user : bool, optional
            Run process as the visiting user (for app content). Default is False.
        default_image_name : str, optional
            Default image for execution if not defined in the bundle. Default is None.
        default_r_environment_management : bool, optional
            Manage R environment for the content. Default is None.
        default_py_environment_management : bool, optional
            Manage Python environment for the content. Default is None.
        service_account_name : str, optional
            Kubernetes service account name for running content. Default is None.
        **attributes : Any
            Additional attributes.

        Returns
        -------
        ContentItem
        """
        path = "v1/content"
        url = self.params.url + path
        response = self.params.session.post(url, json=attrs)
        return ContentItem(self.params, **response.json())

    @overload
    def find(
        self,
        *,
        name: Optional[str] = None,
        owner_guid: Optional[str] = None,
        include: Optional[
            Literal["owner", "tags", "vanity_url"] | list[Literal["owner", "tags", "vanity_url"]]
        ] = None,
    ) -> List[ContentItem]:
        """Find content matching the specified criteria.

        **Applies to Connect versions 2024.06.0 and later.**

        Parameters
        ----------
        name : str, optional
            The content name specified at creation; unique within the owner's account.
        owner_guid : str, optional
            The UUID of the content owner.
        include : str or list of str, optional
            Additional details to include in the response. Allowed values: 'owner', 'tags', 'vanity_url'.

        Returns
        -------
        List[ContentItem]
            List of matching content items.

        Note
        ----
        Specifying both `name` and `owner_guid` returns at most one content item due to uniqueness.
        """

    @overload
    def find(
        self,
        *,
        name: Optional[str] = None,
        owner_guid: Optional[str] = None,
        include: Optional[Literal["owner", "tags"] | list[Literal["owner", "tags"]]] = None,
    ) -> List[ContentItem]:
        """Find content matching the specified criteria.

        **Applies to Connect versions prior to 2024.06.0.**

        Parameters
        ----------
        name : str, optional
            The content name specified at creation; unique within the owner's account.
        owner_guid : str, optional
            The UUID of the content owner.
        include : str or list of str, optional
            Additional details to include in the response. Allowed values: 'owner', 'tags'.

        Returns
        -------
        List[ContentItem]
            List of matching content items.

        Note
        ----
        Specifying both `name` and `owner_guid` returns at most one content item due to uniqueness.
        """

    @overload
    def find(self, include: Optional[str | list[Any]], **conditions) -> List[ContentItem]: ...

    def find(self, include: Optional[str | list[Any]] = None, **conditions) -> List[ContentItem]:
        """Find content matching the specified conditions.

        Returns
        -------
        List[ContentItem]
        """
        if isinstance(include, list):
            include = ",".join(include)

        if include is not None:
            conditions["include"] = include

        if self.owner_guid:
            conditions["owner_guid"] = self.owner_guid

        path = "v1/content"
        url = self.params.url + path
        response = self.params.session.get(url, params=conditions)
        return [
            ContentItem(
                self.params,
                **result,
            )
            for result in response.json()
        ]

    def find_by(
        self,
        **attrs: Unpack[ContentItem._AttrsNotRequired],
    ) -> Optional[ContentItem]:
        """Find the first content record matching the specified attributes.

        There is no implied ordering so if order matters, you should find it yourself.

        Parameters
        ----------
        name : str, optional
            URL-friendly identifier. Allows alphanumeric characters, hyphens ("-"), and underscores ("_").
        title : str, optional
            Content title. Default is None
        description : str, optional
            Content description.
        access_type : Literal['all', 'acl', 'logged_in'], optional
            How content manages viewers.
        owner_guid : str, optional
            The unique identifier of the user who owns this content item.
        connection_timeout : int, optional
            Max seconds without data exchange.
        read_timeout : int, optional
            Max seconds without data received.
        init_timeout : int, optional
            Max startup time for interactive apps.
        idle_timeout : int, optional
            Max idle time before process termination.
        max_processes : int, optional
            Max concurrent processes allowed.
        min_processes : int, optional
            Min concurrent processes required.
        max_conns_per_process : int, optional
            Max client connections per process.
        load_factor : float, optional
            Aggressiveness in spawning new processes (0.0 - 1.0).
        cpu_request : float, optional
            Min CPU units required (1 unit = 1 core).
        cpu_limit : float, optional
            Max CPU units allowed.
        memory_request : int, optional
            Min memory (bytes) required.
        memory_limit : int, optional
            Max memory (bytes) allowed.
        amd_gpu_limit : int, optional
            Number of AMD GPUs allocated.
        nvidia_gpu_limit : int, optional
            Number of NVIDIA GPUs allocated.
        run_as : str, optional
            UNIX user to execute the content.
        run_as_current_user : bool, optional
            Run process as the visiting user (for app content). Default is False.
        default_image_name : str, optional
            Default image for execution if not defined in the bundle.
        default_r_environment_management : bool, optional
            Manage R environment for the content.
        default_py_environment_management : bool, optional
            Manage Python environment for the content.
        service_account_name : str, optional
            Kubernetes service account name for running content.

        Returns
        -------
        Optional[ContentItem]

        Example
        -------
        >>> find_by(name="example-content-name")
        """
        attr_items = attrs.items()
        results = self.find()
        results = (
            result for result in results if all(item in result.items() for item in attr_items)
        )
        return next(results, None)

    @overload
    def find_one(
        self,
        *,
        name: Optional[str] = None,
        owner_guid: Optional[str] = None,
        include: Optional[
            Literal["owner", "tags", "vanity_url"] | list[Literal["owner", "tags", "vanity_url"]]
        ] = None,
    ) -> Optional[ContentItem]:
        """Find first content result matching the specified conditions.

        Parameters
        ----------
        name : str, optional
            The content name specified at creation; unique within the owner's account.
        owner_guid : str, optional
            The UUID of the content owner.
        include : str or list of str, optional
            Additional details to include in the response. Allowed values: 'owner', 'tags', 'vanity_url'.

        Returns
        -------
        Optional[ContentItem]
            List of matching content items.

        Note
        ----
        Specifying both `name` and `owner_guid` returns at most one content item due to uniqueness.
        """

    @overload
    def find_one(
        self,
        *,
        name: Optional[str] = None,
        owner_guid: Optional[str] = None,
        include: Optional[Literal["owner", "tags"] | list[Literal["owner", "tags"]]] = None,
    ) -> Optional[ContentItem]:
        """Find first content result matching the specified conditions.

        **Applies to Connect versions prior to 2024.06.0.**

        Parameters
        ----------
        name : str, optional
            The content name specified at creation; unique within the owner's account.
        owner_guid : str, optional
            The UUID of the content owner.
        include : str or list of str, optional
            Additional details to include in the response. Allowed values: 'owner', 'tags'.

        Returns
        -------
        Optional[ContentItem]
            List of matching content items.

        Note
        ----
        Specifying both `name` and `owner_guid` returns at most one content item due to uniqueness.
        """

    @overload
    def find_one(self, **conditions) -> Optional[ContentItem]: ...

    def find_one(self, **conditions) -> Optional[ContentItem]:
        """Find first content result matching the specified conditions.

        Returns
        -------
        Optional[ContentItem]
        """
        items = self.find(**conditions)
        return next(iter(items), None)

    def get(self, guid: str) -> ContentItem:
        """Get a content item.

        Parameters
        ----------
        guid : str

        Returns
        -------
        ContentItem
        """
        path = f"v1/content/{guid}"
        url = self.params.url + path
        response = self.params.session.get(url)
        return ContentItem(self.params, **response.json())



"""Group resources."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, overload

from .paginator import Paginator
from .resources import Resource, Resources

if TYPE_CHECKING:
    import requests


class Group(Resource):
    def delete(self) -> None:
        """Delete the group."""
        path = f"v1/groups/{self['guid']}"
        url = self.params.url + path
        self.params.session.delete(url)


class Groups(Resources):
    """Groups resource."""

    @overload
    def create(self, *, name: str, unique_id: str | None) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group
        """

    @overload
    def create(self, **kwargs) -> Group:
        """Create a group.

        Returns
        -------
        Group
        """

    def create(self, **kwargs) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group
        """
        path = "v1/groups"
        url = self.params.url + path
        response = self.params.session.post(url, json=kwargs)
        return Group(self.params, **response.json())

    @overload
    def find(
        self,
        *,
        prefix: str = ...,
    ) -> List[Group]: ...

    @overload
    def find(self, **kwargs) -> List[Group]: ...

    def find(self, **kwargs):
        """Find groups.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        List[Group]
        """
        path = "v1/groups"
        url = self.params.url + path
        paginator = Paginator(self.params.session, url, params=kwargs)
        results = paginator.fetch_results()
        return [
            Group(
                self.params,
                **result,
            )
            for result in results
        ]

    @overload
    def find_one(
        self,
        *,
        prefix: str = ...,
    ) -> Group | None: ...

    @overload
    def find_one(self, **kwargs) -> Group | None: ...

    def find_one(self, **kwargs) -> Group | None:
        """Find one group.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        Group | None
        """
        path = "v1/groups"
        url = self.params.url + path
        paginator = Paginator(self.params.session, url, params=kwargs)
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        groups = (
            Group(
                self.params,
                **result,
            )
            for result in results
        )
        return next(groups, None)

    def get(self, guid: str) -> Group:
        """Get group.

        Parameters
        ----------
        guid : str

        Returns
        -------
        Group
        """
        url = self.params.url + f"v1/groups/{guid}"
        response = self.params.session.get(url)
        return Group(
            self.params,
            **response.json(),
        )

    def count(self) -> int:
        """Count the number of groups.

        Returns
        -------
        int
        """
        path = "v1/groups"
        url = self.params.url + path
        response: requests.Response = self.params.session.get(url, params={"page_size": 1})
        result: dict = response.json()
        return result["total"]



from __future__ import annotations

import posixpath
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from ._json import Jsonifiable
    from .context import Context


class ApiCallProtocol(Protocol):
    _ctx: Context
    _path: str

    def _endpoint(self, *path) -> str: ...
    def _get_api(self, *path) -> Jsonifiable: ...
    def _delete_api(self, *path) -> Jsonifiable | None: ...
    def _patch_api(self, *path, json: Jsonifiable | None) -> Jsonifiable: ...
    def _put_api(self, *path, json: Jsonifiable | None) -> Jsonifiable: ...


def endpoint(ctx: Context, *path) -> str:
    return ctx.url + posixpath.join(*path)


# Helper methods for API interactions
def get_api(ctx: Context, *path) -> Jsonifiable:
    response = ctx.session.get(endpoint(ctx, *path))
    return response.json()


def put_api(
    ctx: Context,
    *path,
    json: Jsonifiable | None,
) -> Jsonifiable:
    response = ctx.session.put(endpoint(ctx, *path), json=json)
    return response.json()


# Mixin class for API interactions


class ApiCallMixin:
    def _endpoint(self: ApiCallProtocol, *path) -> str:
        return endpoint(self._ctx, self._path, *path)

    def _get_api(self: ApiCallProtocol, *path) -> Jsonifiable:
        response = self._ctx.session.get(self._endpoint(*path))
        return response.json()

    def _delete_api(self: ApiCallProtocol, *path) -> Jsonifiable | None:
        response = self._ctx.session.delete(self._endpoint(*path))
        if len(response.content) == 0:
            return None
        return response.json()

    def _patch_api(
        self: ApiCallProtocol,
        *path,
        json: Jsonifiable | None,
    ) -> Jsonifiable:
        response = self._ctx.session.patch(self._endpoint(*path), json=json)
        return response.json()

    def _put_api(
        self: ApiCallProtocol,
        *path,
        json: Jsonifiable | None,
    ) -> Jsonifiable:
        response = self._ctx.session.put(self._endpoint(*path), json=json)
        return response.json()



import responses

from posit.connect import Client


class TestOAuthIntegrations:
    @responses.activate
    def test_get_credentials(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:user-session-token",
                        "subject_token": "cit",
                    },
                ),
            ],
            json={
                "access_token": "viewer-token",
                "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                "token_type": "Bearer",
            },
        )
        c = Client(api_key="12345", url="https://connect.example/")
        c._ctx.version = None
        creds = c.oauth.get_credentials("cit")
        assert "access_token" in creds
        assert creds["access_token"] == "viewer-token"



from posit import connect


class TestGroups:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.item = cls.client.groups.create(name="Friends")

    @classmethod
    def teardown_class(cls):
        cls.item.delete()
        assert cls.client.groups.count() == 0

    def test_count(self):
        assert self.client.groups.count() == 1

    def test_get(self):
        assert self.client.groups.get(self.item["guid"])

    def test_find(self):
        assert self.client.groups.find() == [self.item]

    def test_find_one(self):
        assert self.client.groups.find_one() == self.item



"""OAuth association resources."""

from typing import List

from ..resources import Resource, ResourceParameters, Resources


class Association(Resource):
    pass


class IntegrationAssociations(Resources):
    """IntegrationAssociations resource."""

    def __init__(self, params: ResourceParameters, integration_guid: str) -> None:
        super().__init__(params)
        self.integration_guid = integration_guid

    def find(self) -> List[Association]:
        """Find OAuth associations.

        Returns
        -------
        List[Association]
        """
        path = f"v1/oauth/integrations/{self.integration_guid}/associations"
        url = self.params.url + path

        response = self.params.session.get(url)
        return [
            Association(
                self.params,
                **result,
            )
            for result in response.json()
        ]


class ContentItemAssociations(Resources):
    """ContentItemAssociations resource."""

    def __init__(self, params: ResourceParameters, content_guid: str) -> None:
        super().__init__(params)
        self.content_guid = content_guid

    def find(self) -> List[Association]:
        """Find OAuth associations.

        Returns
        -------
        List[Association]
        """
        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        url = self.params.url + path
        response = self.params.session.get(url)
        return [
            Association(
                self.params,
                **result,
            )
            for result in response.json()
        ]

    def delete(self) -> None:
        """Delete integration associations."""
        data = []

        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        url = self.params.url + path
        self.params.session.put(url, json=data)

    def update(self, integration_guid: str) -> None:
        """Set integration associations."""
        data = [{"oauth_integration_guid": integration_guid}]

        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        url = self.params.url + path
        self.params.session.put(url, json=data)



from unittest import mock

import responses

from posit.connect.client import Client
from posit.connect.oauth.associations import Association

from ..api import load_mock, load_mock_list


class TestAssociationAttributes:
    @classmethod
    def setup_class(cls):
        guid = "22644575-a27b-4118-ad06-e24459b05126"
        fake_items = load_mock_list(f"v1/oauth/integrations/{guid}/associations.json")

        assert len(fake_items) == 1
        fake_item = fake_items[0]

        cls.item = Association(mock.Mock(), **fake_item)

    def test_app_guid(self):
        assert self.item["app_guid"] == "f2f37341-e21d-3d80-c698-a935ad614066"

    def test_oauth_integration_guid(self):
        assert self.item["oauth_integration_guid"] == "22644575-a27b-4118-ad06-e24459b05126"

    def test_oauth_integration_name(self):
        assert self.item["oauth_integration_name"] == "keycloak integration"

    def test_oauth_integration_description(self):
        assert self.item["oauth_integration_description"] == "integration description"

    def test_oauth_integration_template(self):
        assert self.item["oauth_integration_template"] == "custom"

    def test_created_time(self):
        assert self.item["created_time"] == "2024-10-01T18:16:09Z"


class TestIntegrationAssociationsFind:
    @responses.activate
    def test(self):
        guid = "22644575-a27b-4118-ad06-e24459b05126"

        # behavior
        mock_get_integration = responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )
        mock_get_association = responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}/associations",
            json=load_mock(f"v1/oauth/integrations/{guid}/associations.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        # invoke
        associations = c.oauth.integrations.get(guid).associations.find()

        assert len(associations) == 1
        association = associations[0]
        assert association["oauth_integration_guid"] == guid

        assert mock_get_integration.call_count == 1
        assert mock_get_association.call_count == 1


class TestContentAssociationsFind:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get_content = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )
        mock_get_association = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}/oauth/integrations/associations",
            json=load_mock(f"v1/content/{guid}/oauth/integrations/associations.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        # invoke
        associations = c.content.get(guid).oauth.associations.find()

        # assert
        assert len(associations) == 1
        association = associations[0]
        assert association["app_guid"] == guid

        assert mock_get_content.call_count == 1
        assert mock_get_association.call_count == 1


class TestContentAssociationsUpdate:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get_content = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        new_integration_guid = "00000000-a27b-4118-ad06-e24459b05126"

        mock_put = responses.put(
            f"https://connect.example/__api__/v1/content/{guid}/oauth/integrations/associations",
            json=[{"oauth_integration_guid": new_integration_guid}],
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        c.content.get(guid).oauth.associations.update(new_integration_guid)

        # assert
        assert mock_put.call_count == 1
        assert mock_get_content.call_count == 1


class TestContentAssociationsDelete:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get_content = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_put = responses.put(
            f"https://connect.example/__api__/v1/content/{guid}/oauth/integrations/associations",
            json=[],
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        c.content.get(guid).oauth.associations.delete()

        # assert
        assert mock_put.call_count == 1
        assert mock_get_content.call_count == 1



from unittest import mock
from unittest.mock import Mock

from posit.connect.groups import Group

from .api import load_mock_dict

session = Mock()
url = Mock()


class TestGroupAttributes:
    @classmethod
    def setup_class(cls):
        guid = "6f300623-1e0c-48e6-a473-ddf630c0c0c3"
        fake_item = load_mock_dict(f"v1/groups/{guid}.json")
        cls.item = Group(mock.Mock(), **fake_item)

    def test_guid(self):
        assert self.item["guid"] == "6f300623-1e0c-48e6-a473-ddf630c0c0c3"

    def test_name(self):
        assert self.item["name"] == "Friends"

    def test_owner_guid(self):
        assert self.item["owner_guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"



from unittest import mock

import requests
import responses
from responses import matchers

from posit.connect.metrics import visits
from posit.connect.resources import ResourceParameters
from posit.connect.urls import Url

from ..api import load_mock, load_mock_dict


class TestVisitAttributes:
    @classmethod
    def setup_class(cls):
        results = load_mock_dict("v1/instrumentation/content/visits?limit=500.json")["results"]
        assert isinstance(results, list)
        first_result_dict = results[0]
        assert isinstance(first_result_dict, dict)
        cls.visit = visits.VisitEvent(
            mock.Mock(),
            **first_result_dict,
        )

    def test_content_guid(self):
        assert self.visit.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"

    def test_user_guid(self):
        assert self.visit.user_guid == "08e3a41d-1f8e-47f2-8855-f05ea3b0d4b2"

    def test_variant_key(self):
        assert self.visit.variant_key == "HidI2Kwq"

    def test_rendering_id(self):
        assert self.visit.rendering_id == 7

    def test_bundle_id(self):
        assert self.visit.bundle_id == 33

    def test_time(self):
        assert self.visit.time == "2018-09-15T18:00:00-05:00"

    def test_data_version(self):
        assert self.visit.data_version == 1

    def test_path(self):
        assert self.visit.path == "/logs"


class TestVisitsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock("v1/instrumentation/content/visits?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock(
                    "v1/instrumentation/content/visits?limit=500&next=23948901087.json"
                ),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
        ]

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))

        # invoke
        events = visits.Visits(params).find()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 1
        assert len(events) == 1


class TestVisitsFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock("v1/instrumentation/content/visits?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock(
                    "v1/instrumentation/content/visits?limit=500&next=23948901087.json"
                ),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
        ]

        # setup
        params = ResourceParameters(requests.Session(), Url("https://connect.example/__api__"))

        # invoke
        event = visits.Visits(params).find_one()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 0
        assert event
        assert event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"


class TestRenameParams:
    def test_start_to_from(self):
        params = {"start": ...}
        params = visits.rename_params(params)
        assert "start" not in params
        assert "from" in params

    def test_end_to_to(self):
        params = {"end": ...}
        params = visits.rename_params(params)
        assert "end" not in params
        assert "to" in params



import json
from typing import Any


class ClientError(Exception):
    def __init__(
        self,
        error_code: int,
        error_message: str,
        http_status: int,
        http_message: str,
        payload: Any = None,
    ):
        self.error_code = error_code
        self.error_message = error_message
        self.http_status = http_status
        self.http_message = http_message
        self.payload = payload
        super().__init__(
            json.dumps(
                {
                    "error_code": error_code,
                    "error_message": error_message,
                    "http_status": http_status,
                    "http_message": http_message,
                    "payload": payload,
                },
            ),
        )



from typing import Optional

from ..client import Client
from .external import is_local

"""
NOTE: The APIs in this module are provided as a convenience and are subject to breaking changes.
"""


class PositAuthenticator:
    def __init__(
        self,
        local_authenticator: Optional[str] = None,
        client: Optional[Client] = None,
        user_session_token: Optional[str] = None,
    ):
        self._local_authenticator = local_authenticator
        self._client = client
        self._user_session_token = user_session_token

    @property
    def authenticator(self) -> Optional[str]:
        if is_local():
            return self._local_authenticator
        return "oauth"

    @property
    def token(self) -> Optional[str]:
        if is_local():
            return None

        # If the user-session-token wasn't provided and we're running on Connect then we raise an exception.
        # user_session_token is required to impersonate the viewer.
        if self._user_session_token is None:
            raise ValueError("The user-session-token is required for viewer authentication.")

        if self._client is None:
            self._client = Client()

        credentials = self._client.oauth.get_credentials(self._user_session_token)
        return credentials.get("access_token")



from __future__ import annotations

from typing import List, overload

from ..cursors import CursorPaginator
from ..resources import Resource, Resources


class ShinyUsageEvent(Resource):
    @property
    def content_guid(self) -> str:
        """The associated unique content identifier.

        Returns
        -------
        str
        """
        return self["content_guid"]

    @property
    def user_guid(self) -> str:
        """The associated unique user identifier.

        Returns
        -------
        str
        """
        return self["user_guid"]

    @property
    def started(self) -> str:
        """The started timestamp.

        Returns
        -------
        str
        """
        return self["started"]

    @property
    def ended(self) -> str:
        """The ended timestamp.

        Returns
        -------
        str
        """
        return self["ended"]

    @property
    def data_version(self) -> int:
        """The data version.

        Returns
        -------
        int
        """
        return self["data_version"]


class ShinyUsage(Resources):
    @overload
    def find(
        self,
        *,
        content_guid: str = ...,
        min_data_version: int = ...,
        start: str = ...,
        end: str = ...,
    ) -> List[ShinyUsageEvent]:
        """Find usage.

        Parameters
        ----------
        content_guid : str, optional
            Filter by an associated unique content identifer, by default ...
        min_data_version : int, optional
            Filter by a minimum data version, by default ...
        start : str, optional
            Filter by the start time, by default ...
        end : str, optional
            Filter by the end time, by default ...

        Returns
        -------
        List[ShinyUsageEvent]
        """

    @overload
    def find(self, **kwargs) -> List[ShinyUsageEvent]:
        """Find usage.

        Returns
        -------
        List[ShinyUsageEvent]
        """

    def find(self, **kwargs) -> List[ShinyUsageEvent]:
        """Find usage.

        Returns
        -------
        List[ShinyUsageEvent]
        """
        params = rename_params(kwargs)

        path = "/v1/instrumentation/shiny/usage"
        url = self.params.url + path
        paginator = CursorPaginator(self.params.session, url, params=params)
        results = paginator.fetch_results()
        return [
            ShinyUsageEvent(
                self.params,
                **result,
            )
            for result in results
        ]

    @overload
    def find_one(
        self,
        *,
        content_guid: str = ...,
        min_data_version: int = ...,
        start: str = ...,
        end: str = ...,
    ) -> ShinyUsageEvent | None:
        """Find a usage event.

        Parameters
        ----------
        content_guid : str, optional
            Filter by an associated unique content identifer, by default ...
        min_data_version : int, optional
            Filter by a minimum data version, by default ...
        start : str, optional
            Filter by the start time, by default ...
        end : str, optional
            Filter by the end time, by default ...

        Returns
        -------
        ShinyUsageEvent | None
        """

    @overload
    def find_one(self, **kwargs) -> ShinyUsageEvent | None:
        """Find a usage event.

        Returns
        -------
        ShinyUsageEvent | None
        """

    def find_one(self, **kwargs) -> ShinyUsageEvent | None:
        """Find a usage event.

        Returns
        -------
        ShinyUsageEvent | None
        """
        params = rename_params(kwargs)
        path = "/v1/instrumentation/shiny/usage"
        url = self.params.url + path
        paginator = CursorPaginator(self.params.session, url, params=params)
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        visits = (
            ShinyUsageEvent(
                self.params,
                **result,
            )
            for result in results
        )
        return next(visits, None)


def rename_params(params: dict) -> dict:
    """Rename params from the internal to the external signature.

    The API accepts `from` as a querystring parameter. Since `from` is a reserved word in Python, the SDK uses the name `start` instead. The querystring parameter `to` takes the same form for consistency.

    Parameters
    ----------
    params : dict

    Returns
    -------
    dict
    """
    if "start" in params:
        params["from"] = params["start"]
        del params["start"]

    if "end" in params:
        params["to"] = params["end"]
        del params["end"]

    return params



import responses
from responses import matchers

from posit.connect.client import Client

from ..api import load_mock


class TestSessionDelete:
    @responses.activate
    def test(self):
        guid = "32c04dc6-0318-41b7-bc74-7e321b196f14"

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/oauth/sessions/{guid}",
            json=load_mock(f"v1/oauth/sessions/{guid}.json"),
        )

        mock_delete = responses.delete(f"https://connect.example/__api__/v1/oauth/sessions/{guid}")

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        session = c.oauth.sessions.get(guid)

        # invoke
        session.delete()

        # assert
        assert mock_delete.call_count == 1


class TestSessionsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/oauth/sessions",
            json=load_mock("v1/oauth/sessions.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        sessions = c.oauth.sessions.find()

        # assert
        assert mock_get.call_count == 1
        assert len(sessions) == 3
        assert sessions[0]["id"] == "54"
        assert sessions[1]["id"] == "55"
        assert sessions[2]["id"] == "56"

    @responses.activate
    def test_params_all(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/oauth/sessions",
            json=load_mock("v1/oauth/sessions.json"),
            match=[matchers.query_param_matcher({"all": True})],
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        c.oauth.sessions.find(all=True)

        # assert
        assert mock_get.call_count == 1


class TestSessionsGet:
    @responses.activate
    def test(self):
        guid = "32c04dc6-0318-41b7-bc74-7e321b196f14"

        # behavior
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/oauth/sessions/{guid}",
            json=load_mock(f"v1/oauth/sessions/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        session = c.oauth.sessions.get(guid=guid)

        # assert
        assert mock_get.call_count == 1
        assert session["guid"] == guid



from posit.connect.resources import ResourceParameters

from .users import User


def get(params: ResourceParameters) -> User:
    """
    Gets the current user.

    Args:
        config (Config): The configuration object containing the URL.
        session (requests.Session): The session object used for making HTTP requests.

    Returns
    -------
        User: The current user.
    """
    url = params.url + "v1/user"
    response = params.session.get(url)
    return User(params, **response.json())



from __future__ import annotations

from typing import Any, Iterator, List, Mapping, MutableMapping, Optional

from .resources import ResourceParameters, Resources


class EnvVars(Resources, MutableMapping[str, Optional[str]]):
    def __init__(self, params: ResourceParameters, content_guid: str) -> None:
        super().__init__(params)
        self.content_guid = content_guid

    def __delitem__(self, key: str, /) -> None:
        """Delete the environment variable.

        Parameters
        ----------
        key : str
            The name of the environment variable to delete.

        Examples
        --------
        >>> vars = EnvVars(params, content_guid)
        >>> del vars["DATABASE_URL"]
        """
        self.update({key: None})

    def __getitem__(self, key: Any) -> Any:
        raise NotImplementedError(
            "Since environment variables may contain sensitive information, the values are not accessible outside of Connect.",
        )

    def __iter__(self) -> Iterator:
        return iter(self.find())

    def __len__(self):
        return len(self.find())

    def __setitem__(self, key: str, value: Optional[str], /) -> None:
        """Set environment variable.

        Set the environment variable for content.

        Parameters
        ----------
        key : str
            The name of the environment variable to set.
        value : str
            The value assigned to the environment variable.

        Examples
        --------
        >>> vars = EnvVars(params, content_guid)
        >>> vars["DATABASE_URL"] = "postgres://user:password@localhost:5432/database"
        """
        self.update({key: value})

    def clear(self) -> None:
        """Remove all environment variables.

        Examples
        --------
        >>> clear()
        """
        path = f"v1/content/{self.content_guid}/environment"
        url = self.params.url + path
        self.params.session.put(url, json=[])

    def create(self, key: str, value: str, /) -> None:
        """Create an environment variable.

        Set an environment variable with the provided key and value. If the key already exists, its value is overwritten without warning to the provided value.

        Parameters
        ----------
        key : str
            The name of the environment variable to create.
        value : str
            The value assigned to the environment variable.

        Examples
        --------
        >>> create(
        ...     "DATABASE_URL",
        ...     "postgres://user:password@localhost:5432/database",
        ... )
        """
        self.update({key: value})

    def delete(self, key: str, /) -> None:
        """Delete the environment variable.

        Parameters
        ----------
        key : str
            The name of the environment variable to delete.

        Examples
        --------
        >>> delete("DATABASE_URL")
        """
        self.update({key: None})

    def find(self) -> List[str]:
        """Find environment variables.

        List the names of the defined environment variables.

        Returns
        -------
        List[str]
            Environment variable names.

        Notes
        -----
        The Connect environment variables API does support retrieving the environment variable's value.

        Examples
        --------
        >>> find()
        ['DATABASE_URL']
        """
        path = f"v1/content/{self.content_guid}/environment"
        url = self.params.url + path
        response = self.params.session.get(url)
        return response.json()

    def items(self):
        raise NotImplementedError(
            "Since environment variables may contain sensitive information, the values are not accessible outside of Connect.",
        )

    def update(self, other=(), /, **kwargs: Optional[str]):
        """
        Update environment variables.

        Updates environment variables with the provided key-value pairs. Accepts a dictionary, an iterable of key-value pairs, or keyword arguments to update the environment variables. All keys and values must be str types.

        Parameters
        ----------
        other : dict, iterable of tuples, optional
            A dictionary or an iterable of key-value pairs to update the environment variables. By default, it is None.
        **kwargs : str
            Additional key-value pairs to update the environment variables.

        Raises
        ------
        TypeError
            If the type of 'other' is not a dictionary or an iterable of key-value pairs.

        Examples
        --------
        Update using keyword arguments:
        >>> update(DATABASE_URL="postgres://user:password@localhost:5432/database")

        Update using multiple keyword arguments:
        >>> update(
        ...     DATABASE_URL="postgres://localhost:5432/database",
        ...     DATABASE_USERNAME="user",
        ...     DATABASE_PASSWORD="password",
        ... )

        Update using a dictionary:
        >>> update(
        ...     {
        ...         "DATABASE_URL": "postgres://localhost:5432/database",
        ...         "DATABASE_USERNAME": "user",
        ...         "DATABASE_PASSWORD": "password",
        ...     }
        ... )

        Update using an iterable of key-value pairs:
        >>> update(
        ...     [
        ...         ("DATABASE_URL", "postgres://localhost:5432/database"),
        ...         ("DATABASE_USERNAME", "user"),
        ...         ("DATABASE_PASSWORD", "password"),
        ...     ]
        ... )
        """
        d = {}
        if isinstance(other, Mapping):
            for key in other:
                d[key] = other[key]
        elif hasattr(other, "keys"):
            for key in other.keys():  # noqa: SIM118
                d[key] = other[key]
        else:
            for key, value in other:
                d[key] = value

        for key, value in kwargs.items():
            d[key] = value

        body = [{"name": key, "value": value} for key, value in d.items()]
        path = f"v1/content/{self.content_guid}/environment"
        url = self.params.url + path
        self.params.session.patch(url, json=body)



import posixpath
from typing import Any, Literal, Optional, overload

from typing_extensions import NotRequired, Required, TypedDict, Unpack

from .context import Context
from .resources import Active, ActiveFinderMethods, ActiveSequence, Resource

JobTag = Literal[
    "unknown",
    "build_report",
    "build_site",
    "build_jupyter",
    "packrat_restore",
    "python_restore",
    "configure_report",
    "run_app",
    "run_api",
    "run_tensorflow",
    "run_python_api",
    "run_dash_app",
    "run_streamlit",
    "run_bokeh_app",
    "run_fastapi_app",
    "run_pyshiny_app",
    "render_shiny",
    "run_voila_app",
    "testing",
    "git",
    "val_py_ext_pkg",
    "val_r_ext_pkg",
    "val_r_install",
]


class Job(Active):
    class _Job(TypedDict):
        # Identifiers
        id: Required[str]
        """A unique identifier for the job."""

        ppid: Required[Optional[str]]
        """Identifier of the parent process."""

        pid: Required[str]
        """Identifier of the process running the job."""

        key: Required[str]
        """A unique key to identify this job."""

        remote_id: Required[Optional[str]]
        """Identifier for off-host execution configurations."""

        app_id: Required[str]
        """Identifier of the parent content associated with the job."""

        variant_id: Required[str]
        """Identifier of the variant responsible for the job."""

        bundle_id: Required[str]
        """Identifier of the content bundle linked to the job."""

        # Timestamps
        start_time: Required[str]
        """RFC3339 timestamp indicating when the job started."""

        end_time: Required[Optional[str]]
        """RFC3339 timestamp indicating when the job finished."""

        last_heartbeat_time: Required[str]
        """RFC3339 timestamp of the last recorded activity for the job."""

        queued_time: Required[Optional[str]]
        """RFC3339 timestamp when the job was added to the queue."""

        # Status and Exit Information
        status: Required[Literal[0, 1, 2]]
        """Current status. Options are 0 (Active), 1 (Finished), and 2 (Finalized)"""

        exit_code: Required[Optional[int]]
        """The job's exit code, available after completion."""

        # Environment Information
        hostname: Required[str]
        """Name of the node processing the job."""

        cluster: Required[Optional[str]]
        """Location where the job runs, either 'Local' or the cluster name."""

        image: Required[Optional[str]]
        """Location of the content in clustered environments."""

        run_as: Required[str]
        """UNIX user responsible for executing the job."""

        # Queue and Scheduling Information
        queue_name: Required[Optional[str]]
        """Name of the queue processing the job, relevant for scheduled reports."""

        # Job Metadata
        tag: Required[JobTag]
        """A tag categorizing the job type. Options are build_jupyter, build_report, build_site, configure_report, git, packrat_restore, python_restore, render_shiny, run_api, run_app, run_bokeh_app, run_dash_app, run_fastapi_app, run_pyshiny_app, run_python_api, run_streamlit, run_tensorflow, run_voila_app, testing, unknown, val_py_ext_pkg, val_r_ext_pkg, and val_r_install."""

    def __init__(self, ctx: Context, path: str, /, **attributes: Unpack[_Job]):
        super().__init__(ctx, path, **attributes)

    def destroy(self) -> None:
        """Destroy the job.

        Submit a request to kill the job.

        Warnings
        --------
        This operation is irreversible.

        Note
        ----
        This action requires administrator, owner, or collaborator privileges.
        """
        endpoint = self._ctx.url + self._path
        self._ctx.session.delete(endpoint)


class Jobs(ActiveFinderMethods[Job], ActiveSequence[Job]):
    def __init__(self, ctx: Context, path: str):
        """A collection of jobs.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions
        path : str
            The HTTP path component for the jobs endpoint (e.g., 'v1/content/544509fc-e4f0-41de-acb4-1fe3a2c1d797/jobs')
        """
        super().__init__(ctx, path, "key")

    def _create_instance(self, path: str, /, **attributes: Any) -> Job:
        """Creates a Job instance.

        Parameters
        ----------
        path : str
            The HTTP path component for the Job resource endpoint (e.g., 'v1/content/544509fc-e4f0-41de-acb4-1fe3a2c1d797/jobs/7add0bc0-0d89-4397-ab51-90ad4bc3f5c9')

        Returns
        -------
        Job
        """
        return Job(self._ctx, path, **attributes)

    class _FindByRequest(TypedDict, total=False):
        # Identifiers
        id: Required[str]
        """A unique identifier for the job."""

        ppid: NotRequired[Optional[str]]
        """Identifier of the parent process."""

        pid: NotRequired[str]
        """Identifier of the process running the job."""

        key: NotRequired[str]
        """A unique key to identify this job."""

        remote_id: NotRequired[Optional[str]]
        """Identifier for off-host execution configurations."""

        app_id: NotRequired[str]
        """Identifier of the parent content associated with the job."""

        variant_id: NotRequired[str]
        """Identifier of the variant responsible for the job."""

        bundle_id: NotRequired[str]
        """Identifier of the content bundle linked to the job."""

        # Timestamps
        start_time: NotRequired[str]
        """RFC3339 timestamp indicating when the job started."""

        end_time: NotRequired[Optional[str]]
        """RFC3339 timestamp indicating when the job finished."""

        last_heartbeat_time: NotRequired[str]
        """RFC3339 timestamp of the last recorded activity for the job."""

        queued_time: NotRequired[Optional[str]]
        """RFC3339 timestamp when the job was added to the queue."""

        # Status and Exit Information
        status: NotRequired[Literal[0, 1, 2]]
        """Current status. Options are 0 (Active), 1 (Finished), and 2 (Finalized)"""

        exit_code: NotRequired[Optional[int]]
        """The job's exit code, available after completion."""

        # Environment Information
        hostname: NotRequired[str]
        """Name of the node processing the job."""

        cluster: NotRequired[Optional[str]]
        """Location where the job runs, either 'Local' or the cluster name."""

        image: NotRequired[Optional[str]]
        """Location of the content in clustered environments."""

        run_as: NotRequired[str]
        """UNIX user responsible for executing the job."""

        # Queue and Scheduling Information
        queue_name: NotRequired[Optional[str]]
        """Name of the queue processing the job, relevant for scheduled reports."""

        # Job Metadata
        tag: NotRequired[JobTag]
        """A tag categorizing the job type. Options are build_jupyter, build_report, build_site, configure_report, git, packrat_restore, python_restore, render_shiny, run_api, run_app, run_bokeh_app, run_dash_app, run_fastapi_app, run_pyshiny_app, run_python_api, run_streamlit, run_tensorflow, run_voila_app, testing, unknown, val_py_ext_pkg, val_r_ext_pkg, and val_r_install."""

    @overload
    def find_by(self, **conditions: Unpack[_FindByRequest]) -> Optional[Job]:
        """Finds the first record matching the specified conditions.

        There is no implied ordering so if order matters, you should specify it yourself.

        Parameters
        ----------
        id : str, not required
            A unique identifier for the job.
        ppid : Optional[str], not required
            Identifier of the parent process.
        pid : str, not required
            Identifier of the process running the job.
        key : str, not required
            A unique key to identify this job.
        remote_id : Optional[str], not required
            Identifier for off-host execution configurations.
        app_id : str, not required
            Identifier of the parent content associated with the job.
        variant_id : str, not required
            Identifier of the variant responsible for the job.
        bundle_id : str, not required
            Identifier of the content bundle linked to the job.
        start_time : str, not required
            RFC3339 timestamp indicating when the job started.
        end_time : Optional[str], not required
            RFC3339 timestamp indicating when the job finished.
        last_heartbeat_time : str, not required
            RFC3339 timestamp of the last recorded activity for the job.
        queued_time : Optional[str], not required
            RFC3339 timestamp when the job was added to the queue.
        status : int, not required
            Current status. Options are 0 (Active), 1 (Finished), and 2 (Finalized)
        exit_code : Optional[int], not required
            The job's exit code, available after completion.
        hostname : str, not required
            Name of the node processing the job.
        cluster : Optional[str], not required
            Location where the job runs, either 'Local' or the cluster name.
        image : Optional[str], not required
            Location of the content in clustered environments.
        run_as : str, not required
            UNIX user responsible for executing the job.
        queue_name : Optional[str], not required
            Name of the queue processing the job, relevant for scheduled reports.
        tag : JobTag, not required
            A tag categorizing the job type. Options are build_jupyter, build_report, build_site, configure_report, git, packrat_restore, python_restore, render_shiny, run_api, run_app, run_bokeh_app, run_dash_app, run_fastapi_app, run_pyshiny_app, run_python_api, run_streamlit, run_tensorflow, run_voila_app, testing, unknown, val_py_ext_pkg, val_r_ext_pkg, and val_r_install.

        Returns
        -------
        Optional[Job]
        """

    @overload
    def find_by(self, **conditions): ...

    def find_by(self, **conditions) -> Optional[Job]:
        return super().find_by(**conditions)


class JobsMixin(Active, Resource):
    """Mixin class to add a jobs attribute to a resource."""

    def __init__(self, ctx, path, /, **attributes):
        """Mixin class which adds a `jobs` attribute to the Active Resource.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions
        path : str
            The HTTP path component for the resource endpoint
        **attributes : dict
            Resource attributes passed
        """
        super().__init__(ctx, path, **attributes)

        path = posixpath.join(path, "jobs")
        self.jobs = Jobs(ctx, path)



import abc
from typing import Callable, Dict, Optional

from ..client import Client
from .external import is_local

"""
NOTE: These APIs are provided as a convenience and are subject to breaking changes:
https://github.com/databricks/databricks-sdk-py#interface-stability
"""

# The Databricks SDK CredentialsProvider == Databricks SQL HeaderFactory
CredentialsProvider = Callable[[], Dict[str, str]]


class CredentialsStrategy(abc.ABC):
    """Maintain compatibility with the Databricks SQL/SDK client libraries.

    https://github.com/databricks/databricks-sql-python/blob/v3.3.0/src/databricks/sql/auth/authenticators.py#L19-L33
    https://github.com/databricks/databricks-sdk-py/blob/v0.29.0/databricks/sdk/credentials_provider.py#L44-L54
    """

    @abc.abstractmethod
    def auth_type(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        raise NotImplementedError


class PositCredentialsProvider:
    def __init__(self, client: Client, user_session_token: str):
        self._client = client
        self._user_session_token = user_session_token

    def __call__(self) -> Dict[str, str]:
        credentials = self._client.oauth.get_credentials(self._user_session_token)
        access_token = credentials.get("access_token")
        if access_token is None:
            raise ValueError("Missing value for field 'access_token' in credentials.")
        return {"Authorization": f"Bearer {access_token}"}


class PositCredentialsStrategy(CredentialsStrategy):
    def __init__(
        self,
        local_strategy: CredentialsStrategy,
        client: Optional[Client] = None,
        user_session_token: Optional[str] = None,
    ):
        self._local_strategy = local_strategy
        self._client = client
        self._user_session_token = user_session_token

    def sql_credentials_provider(self, *args, **kwargs):
        """The sql connector attempts to call the credentials provider w/o any args.

        The SQL client's `ExternalAuthProvider` is not compatible w/ the SDK's implementation of
        `CredentialsProvider`, so create a no-arg lambda that wraps the args defined by the real caller.
        This way we can pass in a databricks `Config` object required by most of the SDK's `CredentialsProvider`
        implementations from where `sql.connect` is called.

        https://github.com/databricks/databricks-sql-python/issues/148#issuecomment-2271561365
        """
        return lambda: self.__call__(*args, **kwargs)

    def auth_type(self) -> str:
        """Returns the auth type currently in use.

        The databricks-sdk client uses the configurated auth_type to create
        a user-agent string which is used for attribution. We should only
        overwrite the auth_type if we are using the PositCredentialsStrategy (non-local),
        otherwise, we should return the auth_type of the configured local_strategy instead
        to avoid breaking someone elses attribution.

        https://github.com/databricks/databricks-sdk-py/blob/v0.29.0/databricks/sdk/config.py#L261-L269

        NOTE: The databricks-sql client does not use auth_type to set the user-agent.
        https://github.com/databricks/databricks-sql-python/blob/v3.3.0/src/databricks/sql/client.py#L214-L219
        """
        if is_local():
            return self._local_strategy.auth_type()
        else:
            return "posit-oauth-integration"

    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        # If the content is not running on Connect then fall back to local_strategy
        if is_local():
            return self._local_strategy(*args, **kwargs)

        # If the user-session-token wasn't provided and we're running on Connect then we raise an exception.
        # user_session_token is required to impersonate the viewer.
        if self._user_session_token is None:
            raise ValueError("The user-session-token is required for viewer authentication.")

        if self._client is None:
            self._client = Client()

        return PositCredentialsProvider(self._client, self._user_session_token)



"""User resources."""

from __future__ import annotations

from typing import List, Literal

from typing_extensions import NotRequired, Required, TypedDict, Unpack

from . import me
from .content import Content
from .paginator import Paginator
from .resources import Resource, ResourceParameters, Resources


class User(Resource):
    @property
    def content(self) -> Content:
        return Content(self.params, owner_guid=self["guid"])

    def lock(self, *, force: bool = False):
        """
        Lock the user account.

        You cannot unlock your own account unless you have administrative privileges. Once an account is locked, only an admin can unlock it.

        Parameters
        ----------
        force : bool, optional
            If `True`, overrides lock protection allowing a user to lock their own account. Default is `False`.

        Returns
        -------
        None

        Examples
        --------
        Lock another user's account:

        >>> user.lock()

        Attempt to lock your own account (will raise `RuntimeError` unless `force` is set to `True`):

        >>> user.lock(force=True)
        """
        _me = me.get(self.params)
        if _me["guid"] == self["guid"] and not force:
            raise RuntimeError(
                "You cannot lock your own account. Set force=True to override this behavior.",
            )
        url = self.params.url + f"v1/users/{self['guid']}/lock"
        body = {"locked": True}
        self.params.session.post(url, json=body)
        super().update(locked=True)

    def unlock(self):
        """
        Unlock the user account.

        This method unlocks the specified user's account. You must have administrative privileges to unlock accounts other than your own.

        Returns
        -------
        None

        Examples
        --------
        Unlock a user's account:

        >>> user.unlock()
        """
        url = self.params.url + f"v1/users/{self['guid']}/lock"
        body = {"locked": False}
        self.params.session.post(url, json=body)
        super().update(locked=False)

    class UpdateUser(TypedDict):
        """Update user request."""

        email: NotRequired[str]
        username: NotRequired[str]
        first_name: NotRequired[str]
        last_name: NotRequired[str]
        user_role: NotRequired[Literal["administrator", "publisher", "viewer"]]

    def update(
        self,
        **kwargs: Unpack[UpdateUser],
    ) -> None:
        """
        Update the user's attributes.

        Parameters
        ----------
        email : str, not required
            The new email address for the user. Default is `None`.
        username : str, not required
            The new username for the user. Default is `None`.
        first_name : str, not required
            The new first name for the user. Default is `None`.
        last_name : str, not required
            The new last name for the user. Default is `None`.
        user_role : Literal["administrator", "publisher", "viewer"], not required
            The new role for the user. Options are `'administrator'`, `'publisher'`, `'viewer'`. Default is `None`.

        Returns
        -------
        None

        Examples
        --------
        Update the user's email and role:

        >>> user.update(email="newemail@example.com", user_role="publisher")

        Update the user's first and last name:

        >>> user.update(first_name="Jane", last_name="Smith")
        """
        url = self.params.url + f"v1/users/{self['guid']}"
        response = self.params.session.put(url, json=kwargs)
        super().update(**response.json())


class Users(Resources):
    """Users resource."""

    def __init__(self, params: ResourceParameters) -> None:
        super().__init__(params)

    class CreateUser(TypedDict):
        """Create user request."""

        username: Required[str]
        # Authentication Information
        password: NotRequired[str]
        user_must_set_password: NotRequired[bool]
        # Profile Information
        email: NotRequired[str]
        first_name: NotRequired[str]
        last_name: NotRequired[str]
        # Role and Permissions
        user_role: NotRequired[Literal["administrator", "publisher", "viewer"]]
        unique_id: NotRequired[str]

    def create(self, **attributes: Unpack[CreateUser]) -> User:
        """
        Create a new user with the specified attributes.

        Applies when server setting 'Authentication.Provider' is set to 'ldap', 'oauth2', 'pam', 'password', 'proxy', or 'saml'.

        Parameters
        ----------
        username : str, required
            The user's desired username.
        password : str, not required
            Applies when server setting 'Authentication.Provider="password"'. Cannot be set when `user_must_set_password` is `True`.
        user_must_set_password : bool, not required
            If `True`, the user is prompted to set their password on first login. When `False`, the `password` parameter is used. Default is `False`. Applies when server setting 'Authentication.Provider="password"'.
        email : str, not required
            The user's email address.
        first_name : str, not required
            The user's first name.
        last_name : str, not required
            The user's last name.
        user_role : Literal["administrator", "publisher", "viewer"], not required
            The user role.  Options are `'administrator'`, `'publisher'`, `'viewer'`. Falls back to server setting 'Authorization.DefaultUserRole'.
        unique_id : str, maybe required
            Required when server is configured with SAML or OAuth2 (non-Google) authentication. Applies when server setting `ProxyAuth.UniqueIdHeader` is set.

        Returns
        -------
        User
            The newly created user.

        Examples
        --------
        Create a user with a predefined password:

        >>> user = client.create(
        ...     username="jdoe",
        ...     email="jdoe@example.com",
        ...     first_name="John",
        ...     last_name="Doe",
        ...     password="s3cur3p@ssword",
        ...     user_role="viewer",
        ... )

        Create a user who must set their own password:

        >>> user = client.create(
        ...     username="jdoe",
        ...     email="jdoe@example.com",
        ...     first_name="John",
        ...     last_name="Doe",
        ...     user_must_set_password=True,
        ...     user_role="viewer",
        ... )
        """
        # todo - use the 'context' module to inspect the 'authentication' object and route to POST (local) or PUT (remote).
        url = self.params.url + "v1/users"
        response = self.params.session.post(url, json=attributes)
        return User(self.params, **response.json())

    class FindUser(TypedDict):
        """Find user request."""

        prefix: NotRequired[str]
        user_role: NotRequired[Literal["administrator", "publisher", "viewer"] | str]
        account_status: NotRequired[Literal["locked", "licensed", "inactive"] | str]

    def find(self, **conditions: Unpack[FindUser]) -> List[User]:
        """
        Find users matching the specified conditions.

        Parameters
        ----------
        prefix : str, not required
            Filter users by prefix (username, first name, or last name). The filter is case-insensitive.
        user_role : Literal["administrator", "publisher", "viewer"], not required
            Filter by user role. Options are `'administrator'`, `'publisher'`, `'viewer'`. Use `'|'` to represent logical OR (e.g., `'viewer|publisher'`).
        account_status : Literal["locked", "licensed", "inactive"], not required
            Filter by account status. Options are `'locked'`, `'licensed'`, `'inactive'`. Use `'|'` to represent logical OR. For example, `'locked|licensed'` includes users who are either locked or licensed.

        Returns
        -------
        List[User]
            A list of users matching the specified conditions.

        Examples
        --------
        Find all users with a username, first name, or last name starting with 'jo':

        >>> users = client.find(prefix="jo")

        Find all users who are either viewers or publishers:

        >>> users = client.find(user_role="viewer|publisher")

        Find all users who are locked or licensed:

        >>> users = client.find(account_status="locked|licensed")
        """
        url = self.params.url + "v1/users"
        paginator = Paginator(self.params.session, url, params={**conditions})
        results = paginator.fetch_results()
        return [
            User(
                self.params,
                **user,
            )
            for user in results
        ]

    def find_one(self, **conditions: Unpack[FindUser]) -> User | None:
        """
        Find a user matching the specified conditions.

        Parameters
        ----------
        prefix : str, optional
            Filter users by prefix (username, first name, or last name). The filter is case-insensitive. Default is `None`.
        user_role : Literal["administrator", "publisher", "viewer"], optional
            Filter by user role. Options are `'administrator'`, `'publisher'`, `'viewer'`. Use `'|'` to represent logical OR (e.g., `'viewer|publisher'`). Default is `None`.
        account_status : Literal["locked", "licensed", "inactive"], optional
            Filter by account status. Options are `'locked'`, `'licensed'`, `'inactive'`. Use `'|'` to represent logical OR. For example, `'locked|licensed'` includes users who are either locked or licensed. Default is `None`.

        Returns
        -------
        User or None
            The first user matching the specified conditions, or `None` if no user is found.

        Examples
        --------
        Find a user with a username, first name, or last name starting with 'jo':

        >>> user = client.find_one(prefix="jo")

        Find a user who is either a viewer or publisher:

        >>> user = client.find_one(user_role="viewer|publisher")

        Find a user who is locked or licensed:

        >>> user = client.find_one(account_status="locked|licensed")
        """
        url = self.params.url + "v1/users"
        paginator = Paginator(self.params.session, url, params={**conditions})
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        users = (
            User(
                self.params,
                **result,
            )
            for result in results
        )
        return next(users, None)

    def get(self, uid: str) -> User:
        """
        Retrieve a user by their unique identifier (guid).

        Parameters
        ----------
        uid : str
            The unique identifier (guid) of the user to retrieve.

        Returns
        -------
        User

        Examples
        --------
        >>> user = client.get("123e4567-e89b-12d3-a456-426614174000")
        """
        url = self.params.url + f"v1/users/{uid}"
        response = self.params.session.get(url)
        return User(
            self.params,
            **response.json(),
        )

    def count(self) -> int:
        """
        Return the total number of users.

        Returns
        -------
        int
        """
        url = self.params.url + "v1/users"
        response = self.params.session.get(url, params={"page_size": 1})
        result: dict = response.json()
        return result["total"]



"""Task resources."""

from __future__ import annotations

from typing import overload

from . import resources


class Task(resources.Resource):
    @property
    def is_finished(self) -> bool:
        """The task state.

        If True, the task has completed. The task may have exited successfully
        or have failed. Inspect the error_code to determine if the task finished
        successfully or not.

        Returns
        -------
        bool
        """
        return self.get("finished", False)

    @property
    def error_code(self) -> int | None:
        """The error code.

        The error code produced by the task. A non-zero value represent an
        error. A zero value represents no error.

        Returns
        -------
        int | None
            Non-zero value indicates an error.
        """
        return self["code"] if self.is_finished else None

    @property
    def error_message(self) -> str | None:
        """The error message.

        Returns
        -------
        str | None
            Human readable error message, or None on success or not finished.
        """
        return self.get("error") if self.is_finished else None

    # CRUD Methods

    @overload
    def update(self, *args, first: int, wait: int, **kwargs) -> None:
        """Update the task.

        Parameters
        ----------
        first : int, default 0
            Line to start output on.
        wait : int, default 0
            Maximum number of seconds to wait for the task to complete.
        """

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the task."""

    def update(self, *args, **kwargs) -> None:
        """Update the task.

        See Also
        --------
        task.wait_for : Wait for the task to complete.

        Notes
        -----
        When waiting for a task to complete, one should consider utilizing `task.wait_for`.

        Examples
        --------
        >>> task.output
        [
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
        ]
        >>> task.update()
        >>> task.output
        [
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
            "Pretium aenean pharetra magna ac placerat vestibulum lectus mauris."
        ]
        """
        params = dict(*args, **kwargs)
        path = f"v1/tasks/{self['id']}"
        url = self.params.url + path
        response = self.params.session.get(url, params=kwargs)
        result = response.json()
        super().update(**result)

    def wait_for(self) -> None:
        """Wait for the task to finish.

        Examples
        --------
        >>> task.wait_for()
        None
        """
        while not self.is_finished:
            self.update()


class Tasks(resources.Resources):
    @overload
    def get(self, *, uid: str, first: int, wait: int) -> Task:
        """Get a task.

        Parameters
        ----------
        uid : str
            Task identifier.
        first : int, default 0
            Line to start output on.
        wait : int, default 0
            Maximum number of seconds to wait for the task to complete.

        Returns
        -------
        Task
        """

    @overload
    def get(self, uid: str, **kwargs) -> Task:
        """Get a task.

        Parameters
        ----------
        uid : str
            Task identifier.

        Returns
        -------
        Task
        """

    def get(self, uid: str, **kwargs) -> Task:
        """Get a task.

        Parameters
        ----------
        uid : str
            Task identifier.

        Returns
        -------
        Task
        """
        path = f"v1/tasks/{uid}"
        url = self.params.url + path
        response = self.params.session.get(url, params=kwargs)
        result = response.json()
        return Task(self.params, **result)



from pathlib import Path

import pytest
from packaging import version

from posit import connect

from . import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION <= version.parse("2023.01.1"),
    reason="Quarto not available",
)
class TestJobs:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(name="example-quarto-minimal")

    @classmethod
    def teardown_class(cls):
        cls.content.delete()
        assert cls.client.content.count() == 0

    def test(self):
        content = self.content

        path = Path("../../../resources/connect/bundles/example-quarto-minimal/bundle.tar.gz")
        path = Path(__file__).parent / path
        path = path.resolve()
        path = str(path)

        bundle = content.bundles.create(path)
        bundle.deploy()

        jobs = content.jobs
        assert len(jobs) == 1

    def test_find_by(self):
        content = self.content

        path = Path("../../../resources/connect/bundles/example-quarto-minimal/bundle.tar.gz")
        path = Path(__file__).parent / path
        path = path.resolve()
        path = str(path)

        bundle = content.bundles.create(path)
        task = bundle.deploy()
        task.wait_for()

        jobs = content.jobs
        assert len(jobs) != 0



from pathlib import Path

import pyjson5 as json

from posit.connect._json import Jsonifiable, JsonifiableDict, JsonifiableList


def load_mock(path: str) -> Jsonifiable:
    """
    Load mock data from a file.

    Reads a JSON or JSONC (JSON with Comments) file and returns the parsed data.

    It's primarily used for loading mock data for tests.

    The file names for mock data should match the query path that they represent.

    Parameters
    ----------
    path : str
        The relative path to the JSONC file.

    Returns
    -------
    Jsonifiable
        The parsed data from the JSONC file.

    Examples
    --------
    >>> data = load_mock("v1/example.json")
    >>> data = load_mock("v1/example.jsonc")
    """
    return json.loads((Path(__file__).parent / "__api__" / path).read_text())


def load_mock_dict(path: str) -> JsonifiableDict:
    result = load_mock(path)
    assert isinstance(result, dict)
    return result


def load_mock_list(path: str) -> JsonifiableList:
    result = load_mock(path)
    assert isinstance(result, list)
    return result


def get_path(path: str) -> Path:
    return Path(__file__).parent / "__api__" / path



from __future__ import annotations

from typing import Optional

from typing_extensions import TypedDict

from ..resources import ResourceParameters, Resources
from .integrations import Integrations
from .sessions import Sessions


class OAuth(Resources):
    def __init__(self, params: ResourceParameters, api_key: str) -> None:
        super().__init__(params)
        self.api_key = api_key

    @property
    def integrations(self):
        return Integrations(self.params)

    @property
    def sessions(self):
        return Sessions(self.params)

    def get_credentials(self, user_session_token: Optional[str] = None) -> Credentials:
        url = self.params.url + "v1/oauth/integrations/credentials"

        # craft a credential exchange request
        data = {}
        data["grant_type"] = "urn:ietf:params:oauth:grant-type:token-exchange"
        data["subject_token_type"] = "urn:posit:connect:user-session-token"
        if user_session_token:
            data["subject_token"] = user_session_token

        response = self.params.session.post(url, data=data)
        return Credentials(**response.json())


class Credentials(TypedDict, total=False):
    access_token: str
    issued_token_type: str
    token_type: str



"""Permission resources."""

from __future__ import annotations

from typing import List, overload

from requests.sessions import Session as Session

from .resources import Resource, ResourceParameters, Resources


class Permission(Resource):
    def delete(self) -> None:
        """Delete the permission."""
        path = f"v1/content/{self['content_guid']}/permissions/{self['id']}"
        url = self.params.url + path
        self.params.session.delete(url)

    @overload
    def update(self, *args, role: str, **kwargs) -> None:
        """Update the permission.

        Parameters
        ----------
        role : str
            The principal role.
        """

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the permission."""

    def update(self, *args, **kwargs) -> None:
        """Update the permission."""
        body = {
            "principal_guid": self.get("principal_guid"),
            "principal_type": self.get("principal_type"),
            "role": self.get("role"),
        }
        body.update(dict(*args))
        body.update(**kwargs)
        path = f"v1/content/{self['content_guid']}/permissions/{self['id']}"
        url = self.params.url + path
        response = self.params.session.put(
            url,
            json=body,
        )
        super().update(**response.json())


class Permissions(Resources):
    def __init__(self, params: ResourceParameters, content_guid: str) -> None:
        super().__init__(params)
        self.content_guid = content_guid

    def count(self) -> int:
        """Count the number of permissions.

        Returns
        -------
        int
        """
        return len(self.find())

    @overload
    def create(self, *, principal_guid: str, principal_type: str, role: str) -> Permission:
        """Create a permission.

        Parameters
        ----------
        principal_guid : str
        principal_type : str
        role : str

        Returns
        -------
        Permission
        """

    @overload
    def create(self, **kwargs) -> Permission:
        """Create a permission.

        Returns
        -------
        Permission
        """

    def create(self, **kwargs) -> Permission:
        """Create a permission.

        Returns
        -------
        Permission
        """
        path = f"v1/content/{self.content_guid}/permissions"
        url = self.params.url + path
        response = self.params.session.post(url, json=kwargs)
        return Permission(params=self.params, **response.json())

    def find(self, **kwargs) -> List[Permission]:
        """Find permissions.

        Returns
        -------
        List[Permission]
        """
        path = f"v1/content/{self.content_guid}/permissions"
        url = self.params.url + path
        response = self.params.session.get(url, json=kwargs)
        results = response.json()
        return [Permission(self.params, **result) for result in results]

    def find_one(self, **kwargs) -> Permission | None:
        """Find a permission.

        Returns
        -------
        Permission | None
        """
        permissions = self.find(**kwargs)
        return next(iter(permissions), None)

    def get(self, uid: str) -> Permission:
        """Get a permission.

        Parameters
        ----------
        uid : str
            The permission id.

        Returns
        -------
        Permission
        """
        path = f"v1/content/{self.content_guid}/permissions/{uid}"
        url = self.params.url + path
        response = self.params.session.get(url)
        return Permission(self.params, **response.json())



import pytest
from packaging import version

from posit import connect
from posit.connect.content import ContentItem, ContentItemRepository

from . import CONNECT_VERSION


class TestContentItemRepository:
    content: ContentItem

    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(name="example")

    @classmethod
    def teardown_class(cls):
        cls.content.delete()
        assert cls.client.content.count() == 0

    @property
    def repo_repository(self):
        return "https://github.com/posit-dev/posit-sdk-py"

    @property
    def repo_branch(self):
        return "1dacc4dd"

    @property
    def repo_directory(self):
        return "integration/resources/connect/bundles/example-quarto-minimal"

    @property
    def repo_polling(self):
        return False

    @property
    def default_repository(self):
        return {
            "repository": self.repo_repository,
            "branch": self.repo_branch,
            "directory": self.repo_directory,
            "polling": self.repo_polling,
        }

    @pytest.mark.skipif(
        # Added to the v2022.12.0 milestone
        # https://github.com/rstudio/connect/issues/22242#event-7859377097
        CONNECT_VERSION < version.parse("2022.12.0"),
        reason="Repository routes not implemented",
    )
    def test_create_get_update_delete(self):
        content = self.content

        # None by default
        assert content.repository is None

        # Create
        new_repo = content.create_repository(**self.default_repository)

        # Get
        content_repo = content.repository
        assert content_repo is not None

        def assert_repo(r: ContentItemRepository):
            assert isinstance(content_repo, ContentItemRepository)
            assert r["repository"] == self.repo_repository
            assert r["branch"] == self.repo_branch
            assert r["directory"] == self.repo_directory
            assert r["polling"] is self.repo_polling

        assert_repo(new_repo)
        assert_repo(content_repo)

        # Update
        ex_branch = "main"
        updated_repo = content_repo.update(branch=ex_branch)
        assert updated_repo["branch"] == ex_branch

        assert updated_repo["repository"] == self.repo_repository
        assert updated_repo["directory"] == self.repo_directory
        assert updated_repo["polling"] is self.repo_polling

        # Delete
        content_repo.destroy()
        assert content.repository is None
